"""
This module acts as a wrapper around some GamryCOM items. If you want to understand it, I highly
recommend going through the example code for the Electrochemistry Toolkit.
"""

import json
import time
import types
from types import ModuleType
from typing import Any, Self

import comtypes.client as client  # type: ignore
from comtypes.client._events import _AdviseConnection  # type: ignore
from PyQt6.QtCore import QObject, pyqtSignal

from ..constants.paths import settings
from ..constants.paths.settings import gamry as Keys


class GamryInterface:
    """Convenience class for interacting with Gamry hardware."""

    def __init__(self) -> None:
        try:
            try:
                with open(settings.gamry.SAVED_SETTINGS_FILE, "r") as f:
                    gamry_data = json.load(f)
            except Exception:
                with open(settings.gamry.DEFAULT_SETTINGS_FILE, "r") as f:
                    gamry_data = json.load(f)
            self.valid: bool = gamry_data[Keys.ENABLED]
            if self.valid:
                gamry_location: str = gamry_data[Keys.LOCATION]
                self.GamryCOM = client.GetModule(gamry_location)
                self.device_list = client.CreateObject(self.GamryCOM.GamryDeviceList)
        except Exception:
            self.valid = False

    def is_valid(self) -> bool:
        """Whether the interface is valid. You should never use an invalid interface."""
        return self.valid

    def com_interface(self) -> types.ModuleType:
        """Get the COM interface used to communicate with GamryCOM."""
        return self.GamryCOM

    def get_pstat_list(self) -> list[str]:
        """Get a list of Gamry potentiostat identifiers."""
        return self.device_list.EnumSections()

    def cleanup(self):
        """Clean up the interface (call this before the application terminates)."""
        try:
            self.device_list.Release()  # closes GamryCOM.exe
        except Exception:
            pass


class Potentiostat:
    """
    Parameters
    ----------
    COM_interface
        The Module object used interface with Gamry.
    identifier
        An identifier for the physical potentiostat.
    """

    def __init__(self, COM_interface: types.ModuleType, identifer: str):
        self.GamryCOM = COM_interface
        self.id = identifer
        self.device = client.CreateObject(self.GamryCOM.GamryPC6Pstat)
        self.device.Init(self.id)

    def identifier(self) -> str:
        """Get the identifier for the potentiostat."""
        return self.id

    def inner(self) -> Any:
        """Access the underlying COM object this potentiostat uses."""
        return self.device

    def com_interface(self) -> ModuleType:
        """Access the potentiostat's underlying COM interface."""
        return self.GamryCOM

    def cleanup(self):
        """Clean up the potentiostat resources."""
        self.close()

    def initialize(self, dc_voltage_vs_reference: float) -> Self:
        """
        Initialize the potentiostat the same way Gamry does for the Potentiostatic EIS experiment.
        """
        # I know it's ugly, I'm sorry my child
        self.turn_on()
        self.device.SetCtrlMode(self.GamryCOM.PstatMode)
        self.device.SetAchSelect(self.GamryCOM.GND)
        self.device.SetIEStability(self.GamryCOM.StabilityFast)
        self.device.SetCASpeed(3)
        self.device.SetSenseSpeedMode(True)
        self.device.SetGround(self.GamryCOM.Float)
        self.device.SetIchRange(3.0)
        self.device.SetIchRangeMode(False)
        self.device.SetIchFilter(2.5)
        self.device.SetVchRange(3.0)
        self.device.SetVchRangeMode(False)

        self.device.SetIchOffsetEnable(True)
        self.device.SetVchOffsetEnable(True)

        self.device.SetVchFilter(2.5)
        self.device.SetAchRange(3.0)
        self.device.SetIERangeLowerLimit(0)
        self.device.SetIERange(0.03)
        self.device.SetIERangeMode(False)
        self.device.SetAnalogOut(0.0)
        self.device.SetVoltage(dc_voltage_vs_reference)
        self.device.SetPosFeedEnable(False)
        self.device.SetIruptMode(self.GamryCOM.IruptOff)

        return self

    def upper_frequency_limit(self) -> float:
        """Get the upper limit for the potentiostat's frequency."""
        return self.device.FreqLimitUpper()

    def lower_frequency_limit(self) -> float:
        """Get the lower limit for the potentiostat's frequency."""
        return self.device.FreqLimitLower()

    def clamp_frequency(self, frequency: float):
        """
        Clamp the input frequency to be between the upper and lower limit for the potentiostat,
        then return the resulting frequency.
        """
        lower_limit = self.lower_frequency_limit()
        upper_limit = self.upper_frequency_limit()
        if frequency > upper_limit:
            return upper_limit
        elif frequency < lower_limit:
            return lower_limit
        return frequency

    def open(self) -> Self:
        """
        Open to potentiostat for taking measurements. This must eventually be followed by a call to
        `close()`.
        """
        self.device.Open()
        return self

    def close(self) -> Self:
        """Close the potentiostat. You cannot take measurements after calling this function."""
        self.turn_off()
        time.sleep(0.5)  # necessary to make sure the potentiostat actually turns off
        self.device.Close()
        return self

    def turn_on(self) -> Self:
        """Turn the cell on."""
        self.device.SetCell(self.GamryCOM.CellOn)
        return self

    def turn_off(self) -> Self:
        """Turn the cell off."""
        self.device.SetCell(self.GamryCOM.CellOff)
        return self

    def measure_open_circuit_voltage(self) -> float:
        """Measure the open-circuit voltage."""
        voltage_reader = OCVoltageReader(self)
        return voltage_reader.run()

    # ----------------------------------------------------------------------------------------------
    # context manager
    def __enter__(self) -> Self:
        self.open()
        return self

    def __exit__(self, *exc_args):
        self.cleanup()


class ImpedanceReader(QObject):
    """
    Uses a `Potentiostat` to measure electrical impedance. This is synonymous to Gamry's
    `ReadZ`.
    """

    dataReady = pyqtSignal(bool)
    """
    Fires sometime after a call to `measure()` when the impedance reader has data ready. Sends:
    - Whether the measurement was successful as a `bool`.
    """

    def __init__(self, potentiostat: Potentiostat):
        QObject.__init__(self)
        self.pstat = potentiostat
        # stands for Read Z (Z = impedance)
        self.readz = client.CreateObject(self.pstat.com_interface().GamryReadZ)
        self.readz.Init(self.pstat.inner())
        self.connection = client.GetEvents(self.readz, self)

    def initialize(self, impedance_guess: float, speed_option: int) -> Self:
        """Initialize the reader the same way Gamry does in the Potentiostatic EIS experiment."""
        self.readz.SetSpeed(speed_option)
        self.readz.SetGain(1.0)
        self.readz.SetINoise(0.0)
        self.readz.SetVNoise(0.0)
        self.readz.SetIENoise(0.0)
        self.readz.SetZmod(impedance_guess)
        self.readz.SetIdc(self.pstat.inner().MeasureI())
        return self

    def measure(self, frequency: float, ac_voltage: float):
        """
        Perform a measurement. The reader will emit `dataReady` when data is ready to be read.

        Parameters
        ----------
        frequency
            The frequency to measure at (in Hz). This frequency is clamped to the potentiostats
            limits.
        ac_voltage
            The AC voltage to measure at (in V).
        """
        self.readz.Measure(self.pstat.clamp_frequency(frequency), ac_voltage)

    def cleanup(self):
        """Clean up the reader resources."""
        self.connection.disconnect()
        del self.connection

    def potentiostat(self) -> Potentiostat:
        """Access the potentiostat this reader is using."""
        return self.pstat

    def com_connection(self) -> _AdviseConnection:
        """Get the `comtypes` connection used to send events to the reader."""
        return self.connection

    # ----------------------------------------------------------------------------------------------
    # measurement values
    def frequency(self) -> float:
        """Get the current frequency in Hz."""
        return self.readz.Zfreq()

    def real_impedance(self) -> float:
        """Get the real part of the current impedance in Ohms."""
        return self.readz.Zreal()

    def imaginary_impedance(self) -> float:
        """Get the imaginary part of the current impedance in Ohms."""
        return self.readz.Zimag()

    def impedance_standard_deviation(self) -> float:
        """Get the current impedance standard deviation. This is also called Zsig."""
        return self.readz.Zsig()

    def impedance_magnitude(self) -> float:
        """Get the magnitude of the current impedance in Ohms. This is also called Zmod."""
        return self.readz.Zmod()

    def impedance_phase(self) -> float:
        """Get the phase of the current impedance in degrees."""
        return self.readz.Zphz()

    def dc_current(self) -> float:
        """Get the current DC current in Amps."""
        return self.readz.Idc()

    def dc_voltage(self) -> float:
        """Get the current DC voltage in Volts."""
        return self.readz.Vdc()

    def ie_range(self) -> float:
        """Get the current IE range (also called current range)."""
        return self.readz.IERange()

    # ----------------------------------------------------------------------------------------------
    # COM functions
    def _IGamryReadZEvents_OnDataDone(self, this: Any, error_status: bool):
        """
        This is a callback called by the `comtypes` module when the a potentiostat has completed a
        measurement.
        """
        self.dataReady.emit(error_status == 0)  # 0 indicates success

    def _IGamryReadZEvents_OnDataAvailable(self, this: Any):
        """Another callback that the application does not use right now."""
        pass

    # ----------------------------------------------------------------------------------------------
    # context manager
    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_args):
        self.cleanup()


class OCVoltageReader:
    """Uses a `Potentiostat` to measure the open-circuit voltage."""

    SAMPLE_TIME = 0.25  # seconds

    def __init__(self, potentiostat: Potentiostat):
        self.pstat = potentiostat
        gamry_pstat = self.pstat.inner()
        self.GamryCOM = self.pstat.com_interface()
        self.signal = client.CreateObject(self.GamryCOM.GamrySignalConst)
        self.oc_reader = client.CreateObject(self.GamryCOM.GamryDtaqOcv)
        self.signal.Init(gamry_pstat, 0, 1, self.SAMPLE_TIME, self.GamryCOM.PstatMode)
        self.oc_reader.Init(gamry_pstat)

        self.finished = False
        self.open_circuit_voltage = 0.0

    def inner(self) -> Any:
        """Get the underlying COM object used to read the open-circuit voltage."""
        return self.oc_reader

    def run(self) -> float:
        """
        Run the open-circuit voltage test. This turns the cell off.

        Returns
        -------
        The open-circuit voltage in Volts.
        """
        # a lot of this code is probably redundant, but I'm copying from the Common Functions.exp
        # file (OCDelay)
        gamry_pstat = self.pstat.inner()
        gamry_pstat.SetCell(self.GamryCOM.CellOff)
        gamry_pstat.SetCtrlMode(self.GamryCOM.PstatMode)

        gamry_pstat.SetSenseSpeedMode(True)
        gamry_pstat.SetIConvention(self.GamryCOM.Anodic)
        gamry_pstat.SetGround(self.GamryCOM.Float)
        gamry_pstat.SetIchRange(3.0)
        gamry_pstat.SetIchRangeMode(True)
        gamry_pstat.SetIchOffsetEnable(False)
        gamry_pstat.SetIchFilter(1.0 / self.SAMPLE_TIME)
        gamry_pstat.SetVchRange(10.0)
        gamry_pstat.SetVchRangeMode(True)
        gamry_pstat.SetVchOffsetEnable(False)
        gamry_pstat.SetVchFilter(1.0 / self.SAMPLE_TIME)
        gamry_pstat.SetAchRange(3.0)
        gamry_pstat.SetIERange(0.03)
        gamry_pstat.SetIERangeMode(False)
        gamry_pstat.SetAnalogOut(0.0)
        gamry_pstat.SetVoltage(0.0)
        gamry_pstat.SetPosFeedEnable(False)
        gamry_pstat.SetIruptMode(self.GamryCOM.IruptOff)

        gamry_pstat.SetSignal(self.signal)
        gamry_pstat.InitSignal()

        gamry_pstat.FindVchRange()

        # actually run the open-circuit voltage test
        connection = client.GetEvents(self.oc_reader, self)
        self.oc_reader.Run(True)
        while not self.finished:
            client.PumpEvents(0.1)
        connection.disconnect()
        del connection

        return self.open_circuit_voltage  # return the last measured open-circuit voltage

    # ----------------------------------------------------------------------------------------------
    # COM functions
    def _IGamryDtaqEvents_OnDataAvailable(self, this) -> None:
        """Called whenever there are new open circuit voltage measurements."""
        points: list[list[float]] = self.oc_reader.Cook(1024)[1]
        self.open_circuit_voltage = points[1][-1]  # Vf

    def _IGamryDtaqEvents_OnDataDone(self, this):
        """Gets called when the timeout is reached."""
        self.finished = True


GAMRY = GamryInterface()



from fabrial.utility import plugins

from .test_modules import bad_module, good_module

GOOD_PLUGIN_COUNT = 2
BAD_PLUGIN_COUNT = 1
BAD_PLUGIN_FAILURE_COUNT = 1


def test_load_plugins_from_module():
    """Tests `plugins.load_plugins_from_module()`."""
    # good module, we should see all modules loaded and no errors
    plugin_modules, failure_plugins = plugins.load_plugins_from_module(good_module)
    assert len(plugin_modules) == GOOD_PLUGIN_COUNT
    assert len(failure_plugins) == 0  # we should have no failures
    assert plugin_modules["normal1"] == good_module.normal1
    assert plugin_modules["normal2"] == good_module.normal2

    # bad module, we should see a failure
    plugin_modules, failure_plugins = plugins.load_plugins_from_module(bad_module)
    assert len(plugin_modules) == BAD_PLUGIN_COUNT
    assert len(failure_plugins) == BAD_PLUGIN_FAILURE_COUNT
    assert plugin_modules["normal"] == bad_module.normal



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget

from ..instrument_connection import InstrumentConnectionWidget
from ..passive_monitoring import PassiveMonitoringWidget
from ..setpoint import SetpointWidget
from ..stability_check import StabilityCheckWidget
from ..utility import layout as layout_util


class OvenControlTab(QWidget):
    """First tab in the application, used for directly controlling the oven."""

    ICON_FILE = "thermometer.png"

    def __init__(self) -> None:
        # data members
        self.setpoint_widget: SetpointWidget
        self.passive_monitoring_widget: PassiveMonitoringWidget
        self.instrument_connection_widget: InstrumentConnectionWidget
        self.stability_widget: StabilityCheckWidget

        QWidget.__init__(self)

        self.create_widgets()

    def create_widgets(self):
        """Create subwidgets."""
        # create the layout
        layout = QVBoxLayout()
        top_layout = layout_util.add_sublayout(layout, QHBoxLayout())
        self.setLayout(layout)

        self.setpoint_widget = SetpointWidget()
        self.stability_widget = StabilityCheckWidget()
        # do not move these two above the other ones
        self.passive_monitoring_widget = PassiveMonitoringWidget()
        self.instrument_connection_widget = InstrumentConnectionWidget()
        # add the widgets
        layout_util.add_to_layout(
            top_layout,
            self.setpoint_widget,
            self.passive_monitoring_widget,
            self.instrument_connection_widget,
        )
        layout.addWidget(self.stability_widget, alignment=Qt.AlignmentFlag.AlignHCenter)



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QApplication, QMenu, QMenuBar

from ..classes import Action
from ..custom_widgets import OkCancelDialog

if TYPE_CHECKING:
    from ..main_window import MainWindow


class FileMenu(QMenu):
    """File menu."""

    def __init__(self, parent: QMenuBar, main_window: MainWindow):
        QMenu.__init__(self, "&File", parent)
        # silly
        self.addAction(Action(parent, "Honk", lambda: OkCancelDialog("Honk", "Honk").exec()))

        self.addSeparator()

        self.addAction(Action(parent, "Close Window", main_window.close, shortcut="Alt+F4"))
        self.addAction(Action(parent, "Exit", QApplication.closeAllWindows))



from types import ModuleType
from typing import Mapping, Self

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QKeyEvent
from PyQt6.QtWidgets import QSizePolicy, QVBoxLayout

from ...constants.paths.settings import sequence as sequence_paths
from ...custom_widgets import Container, FixedButton
from ...utility import layout as layout_util
from ..tree_models import OptionsModel
from .tree_view import TreeView


class OptionsTreeView(TreeView[OptionsModel]):
    """`TreeView` containing item options for the sequence."""

    @classmethod
    def from_plugins(cls, plugin_modules: Mapping[str, ModuleType]) -> Self:
        """Create from the application's available plugins."""
        view = cls(OptionsModel.from_plugins(plugin_modules))
        view.init_view_state_from_json(sequence_paths.OPTIONS_STATE_FILE)
        return view

    def save_on_close(self):
        """Call this when the application closes to save the view state."""
        self.save_view_state_to_json(sequence_paths.OPTIONS_STATE_FILE)

    def keyPressEvent(self, event: QKeyEvent | None):  # overridden
        if event is not None:
            match event.key():
                case Qt.Key.Key_Return | Qt.Key.Key_Enter:
                    self.open_event(self.selectedIndexes())
        TreeView.keyPressEvent(self, event)


class OptionsTreeWidget(Container):
    """OptionsTreeView with a button for expanding and un-expanding all items."""

    def __init__(self, view: OptionsTreeView):
        layout = QVBoxLayout()
        Container.__init__(self, layout)
        self.view = view
        self.expand_button = FixedButton("Toggle Expansion")
        self.expand_button.setCheckable(True)
        self.expand_button.toggled.connect(
            lambda checked: self.view.expandAll() if checked else self.view.collapseAll()
        )
        layout_util.add_to_layout(layout, self.expand_button, self.view)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

    @classmethod
    def from_plugins(cls, plugin_modules: Mapping[str, ModuleType]) -> Self:
        """Create from the application's available plugins."""
        return cls(OptionsTreeView.from_plugins(plugin_modules))

    def save_on_close(self):
        """Call this when closing the application to save settings."""
        self.view.save_on_close()



from pathlib import Path

from fabrial import FilesDescription, ItemWidget, Substitutions
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QDoubleSpinBox, QFormLayout

# we'll refer to these constants a few times
NAME = "Random Data"
DIRECTORY = Path(__file__).parent
# you should create the icon once instead of inside `__init__()`
ICON = QIcon(str(DIRECTORY.joinpath("example_icon")))


# we inherit from `ItemWidget` so Fabrial can work with our custom widget
class RandomDataWidget(ItemWidget):
    """Record random data on an interval; widget."""

    def __init__(self, data_interval: float):
        # this is a layout for other widgets
        layout = QFormLayout()
        # you must call the base `__init__()` method
        # for now, we're not supplying a description
        ItemWidget.__init__(
            self,
            layout,
            NAME,
            ICON,
            FilesDescription(
                # we provide the absolute path to the descriptions folder
                DIRECTORY.joinpath("descriptions"),
                NAME,
                Substitutions(
                    # notice how the dictionary keys correspond to the
                    # substitutions in the files
                    parameters={"DATA_INTERVAL": "Data Interval"},
                    data_recording={"DATA_FILE": "random_data.txt"},
                ),
            ),
        )

        # let's create the widgets!
        self.interval_spinbox = QDoubleSpinBox()
        # minimum and maximum values
        self.interval_spinbox.setMinimum(0.1)
        self.interval_spinbox.setMaximum(1000)
        # start at the supplied value
        self.interval_spinbox.setValue(data_interval)
        # add the spinbox to the layout
        # this creates an entry labeled "Data Interval" where the user
        # can enter the desired data interval
        layout.addRow("Data Interval", self.interval_spinbox)



import csv
import time
from io import TextIOWrapper

from ..constants.paths.process import headers
from .datetime import get_datetime


def create_temperature_file(filepath: str) -> TextIOWrapper:
    """
    Create a CSV file that records data in the following format:

    [seconds since start],[datetime],[temperature (degrees C)]

    The header is written automatically.

    Parameters
    ----------
    filepath
        The path to the file to create. If the file already exists, it will be overwritten. The file
        is opened in write mode and is line buffered.
    """
    file = open(filepath, "w", 1, newline="")
    csv.writer(file).writerow(
        [headers.time.TIME, headers.time.TIME_DATETIME, headers.oven.TEMPERATURE]
    )
    return file


def record_temperature_data(
    file: TextIOWrapper, start_time: float, temperature: float | None
) -> float:
    """
    Record the time since start, datetime, and temperature in the provided file.

    Parameters
    ----------
    file
        The CSV file to write to.
    start_time
        The `Process`'s start time.
    temperature
        The temperature being recorded.

    Returns
    -------
    The number of seconds since the `Process` started.
    """
    current_time = time.time()
    time_since_start = current_time - start_time
    csv.writer(file).writerow([time_since_start, get_datetime(current_time), str(temperature)])
    return time_since_start



from typing import Any

from PyQt6.QtWidgets import QFormLayout

from .....classes import DescriptionInfo
from .....custom_widgets import SpinBox
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import LoopProcess


class LoopWidget(AbstractBaseWidget):
    DISPLAY_NAME_PREFIX = "Loop"

    def __init__(self):
        layout = QFormLayout()
        AbstractBaseWidget.__init__(
            self,
            layout,
            self.DISPLAY_NAME_PREFIX,
            LoopProcess,
            "arrow-repeat.png",
            DescriptionInfo(
                "flow_control",
                "loop",
                LoopProcess.directory_name(),
                DescriptionInfo.Substitutions(
                    parameters_dict={"NUM_LOOPS": encoding.NUMBER_OF_LOOPS}
                ),
            ),
            True,
        )

        self.loop_spinbox = SpinBox()
        self.loop_spinbox.setMinimum(1)
        layout.addRow(encoding.NUMBER_OF_LOOPS, self.loop_spinbox)
        self.loop_spinbox.textChanged.connect(
            lambda value_as_str: self.setWindowTitle(f"{self.DISPLAY_NAME_PREFIX} ({value_as_str})")
        )

    @classmethod
    def from_dict(cls, data_as_dict: dict[str, Any]):
        widget = cls()
        widget.loop_spinbox.setValue(data_as_dict[encoding.NUMBER_OF_LOOPS])
        return widget

    def to_dict(self) -> dict[str, Any]:
        return {encoding.NUMBER_OF_LOOPS: self.loop_spinbox.value()}



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from types import ModuleType
from typing import Iterable, Mapping, MutableMapping

from ..sequence_builder import CategoryItem, DataItem, SequenceItem, TreeItem


@dataclass
class CategoryInfo:
    """Container used internally by the application. This is not for plugins."""

    items: list[SequenceItem]
    subcategories: dict[str, CategoryInfo]


@dataclass
class PluginCategory:
    """Container that represents an item category from a plugin."""

    name: str
    items: Iterable[DataItem]
    subcategories: Iterable[PluginCategory] = field(default_factory=list)


def items_from_plugins(
    plugin_modules: Mapping[str, ModuleType],
) -> tuple[list[CategoryItem], list[str], list[str]]:
    """
    Helper function for `OptionsModel.from_plugins()`.

    Creates a list of `CategoryItem`s from a group of plugins. Logs errors.

    Returns
    -------
    A tuple of (the created `CategoryItem`s, any plugins that couldn't be loaded, any categories
    that couldn't be loaded).
    """

    failure_plugins: list[str] = []
    failure_categories: list[str] = []

    # helper function to parse plugin categories into the `category_info_map`. This function can
    # fail, so it must be wrapped in a try/except block
    def parse_plugin_categories(
        plugin_categories: Iterable[PluginCategory],
        category_info_map: MutableMapping[str, CategoryInfo],
    ):
        for plugin_category in plugin_categories:
            # get or create the combined category
            try:
                category_info = category_info_map[plugin_category.name]
            except KeyError:
                category_info = CategoryInfo([], {})
                category_info_map[plugin_category.name] = category_info
            # add the plugin's items (after building them into `SequenceItem`s)
            category_info.items.extend(
                [SequenceItem(None, data_item) for data_item in plugin_category.items]
            )
            # parse subcategories (recursive)
            parse_plugin_categories(plugin_category.subcategories, category_info.subcategories)

    # helper function to parse the category info map into actual `CategoryItem`s
    def parse_into_items(category_info_map: dict[str, CategoryInfo]) -> list[CategoryItem]:
        category_items: list[CategoryItem] = []
        for category_name, category_info in sorted(category_info_map.items()):
            try:
                # get the subcategory items (they will already be sorted)
                sub_category_items = parse_into_items(category_info.subcategories)  # recurse
                # sort the sequence items
                category_info.items.sort(key=lambda item: item.display_name())
                # combine the subcategory items and sequence items
                items: list[TreeItem] = []
                items.extend(sub_category_items)
                items.extend(category_info.items)
                # create and append the `CategoryItem`
                category_items.append(CategoryItem(None, category_name, items))
            except Exception:
                logging.getLogger(__name__).exception(
                    "Exception while parsing `CategoryInfo`s into `CategoryItem`s"
                )
                failure_categories.append(category_name)

        return category_items

    category_info_map: dict[str, CategoryInfo] = {}
    for name, plugin_module in plugin_modules.items():
        try:
            # PLUGIN.categories() is a mandatory entry point for plugins
            plugin_categories: list[PluginCategory] = plugin_module.categories()
            parse_plugin_categories(plugin_categories, category_info_map)
        except Exception:
            logging.getLogger(__name__).exception(
                "Exception while parsing plugin categories into `CategoryInfo`s"
            )
            failure_plugins.append(name)

    return (parse_into_items(category_info_map), failure_plugins, failure_categories)



from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QLayout, QSizePolicy, QWidget


class SignalWidget(QWidget):
    """`QWidget` with a `closed` signal."""

    closed = pyqtSignal()

    def __init__(self):
        QWidget.__init__(self)

    def closeEvent(self, event):  # overridden method
        if event is not None:
            self.closed.emit()
        QWidget.closeEvent(self, event)


class FixedWidget(QWidget):
    """`QWidget` with a fixed size."""

    def __init__(self):
        QWidget.__init__(self)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)


class Widget(QWidget):
    """
    `QWidget` that automatically sets the layout.

    Parameters
    ----------
    layout
        The layout to initialize with.
    """

    def __init__(self, layout: QLayout | None):
        QWidget.__init__(self)
        if layout is not None:
            self.setLayout(layout)



"""Uses a fake oven to simulate reading temperatures."""

import time

from PyQt6.QtWidgets import QHBoxLayout, QPushButton, QVBoxLayout

from fabrial import MainWindow
from fabrial.custom_widgets import GroupBox, TemperatureSpinBox
from fabrial.instruments import INSTRUMENTS, ConnectionStatus, Oven
from fabrial.main import main
from fabrial.utility import layout as layout_util


class DeveloperOven(Oven):
    def __init__(self):
        Oven.__init__(self)
        self.unlocked = True
        self.developer_temperature = 0.0
        self.developer_setpoint = 0.0
        self.set_connected(ConnectionStatus.CONNECTED)
        self.set_unlocked(True)

    def set_connected(self, connection_status: ConnectionStatus):
        """Set the connection status and send signals."""
        self.update_connection_status(connection_status)

    def set_unlocked(self, unlocked: bool):
        """Set the unlocked status without firing signals."""
        self.unlocked = unlocked

    def set_temperature(self, temperature: float):
        """Set the oven's temperature."""
        self.developer_temperature = temperature

    # overridden methods

    # no need to override acquire() or release()

    # no need to override is_connected()

    def update_connection_status(self, connection_status: ConnectionStatus):
        self.connection_status = connection_status
        self.connectionChanged.emit(self.is_connected())

    def is_unlocked(self):
        """Return the previously set unlocked status."""
        return self.unlocked

    def read_temp(self) -> float | None:
        """Return the previously set temperature if connected, else None."""
        if self.is_connected():
            self.update_temperature(self.developer_temperature)
            return self.developer_temperature
        else:
            self.update_connection_status(ConnectionStatus.DISCONNECTED)
            return None

    def change_setpoint(self, setpoint):
        """Set the setpoint."""
        self.developer_setpoint = setpoint
        if not self.is_connected():
            self.update_connection_status(ConnectionStatus.DISCONNECTED)
        else:
            self.update_setpoint(setpoint)
            self.reset_stability()
        return self.is_connected()

    def get_setpoint(self) -> float | None:
        if self.is_connected():
            self.update_setpoint(self.developer_setpoint)
            return self.developer_setpoint
        else:
            self.update_connection_status(ConnectionStatus.DISCONNECTED)
        return None

    def connect(self):
        return

    def set_port(self, port):
        """Update the port variable and do nothing else."""
        self.port = port
        pass


class DeveloperWidget(GroupBox):
    def __init__(self):
        layout = QVBoxLayout()
        GroupBox.__init__(self, "Developer", layout)
        self.oven = INSTRUMENTS.oven
        self.create_connection_widgets(layout)
        self.create_connection_signal_widgets(layout)
        self.create_lock_widgets(layout)
        self.create_lock_signal_widgets(layout)
        self.create_temperature_widgets(layout)
        self.create_setpoint_widgets(layout)
        self.create_freeze_widgets(layout)

    def create_connection_widgets(self, layout):
        connect_button = QPushButton("Connect")
        connect_button.pressed.connect(lambda: self.oven.set_connected(ConnectionStatus.CONNECTED))
        disconnect_button = QPushButton("Disconnect")
        disconnect_button.pressed.connect(
            lambda: self.oven.set_connected(ConnectionStatus.DISCONNECTED)
        )
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), connect_button, disconnect_button
        )

    def create_connection_signal_widgets(self, layout):
        connected_button = QPushButton("Emit Connected")
        connected_button.pressed.connect(lambda: self.oven.connectionChanged.emit(True))
        disconnected_button = QPushButton("Emit Disconnect")
        disconnected_button.pressed.connect(lambda: self.oven.connectionChanged.emit(False))
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), connected_button, disconnected_button
        )

    def create_lock_widgets(self, layout):
        lock_button = QPushButton("Lock")
        lock_button.pressed.connect(lambda: self.oven.set_unlocked(False))
        unlock_button = QPushButton("Unlock")
        unlock_button.pressed.connect(lambda: self.oven.set_unlocked(True))
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), unlock_button, lock_button
        )

    def create_lock_signal_widgets(self, layout):
        acquire_button = QPushButton("Emit Acquired")
        acquire_button.pressed.connect(lambda: self.oven.lockChanged.emit(False))
        release_button = QPushButton("Emit Released")
        release_button.pressed.connect(lambda: self.oven.lockChanged.emit(True))
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), release_button, acquire_button
        )

    def create_temperature_widgets(self, layout):
        temperature_spinbox = TemperatureSpinBox(INSTRUMENTS.oven)
        temperature_spinbox.lineEdit().returnPressed.connect(
            lambda: self.oven.set_temperature(temperature_spinbox.value())
        )
        temperature_button = QPushButton("Change Temperature")
        temperature_button.pressed.connect(
            lambda: self.oven.set_temperature(temperature_spinbox.value())
        )
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), temperature_spinbox, temperature_button
        )

    def create_setpoint_widgets(self, layout):
        setpoint_spinbox = TemperatureSpinBox(INSTRUMENTS.oven)
        setpoint_spinbox.lineEdit().returnPressed.connect(
            lambda: self.oven.change_setpoint(setpoint_spinbox.value())
        )
        setpoint_button = QPushButton("Change Setpoint")
        setpoint_button.pressed.connect(lambda: self.oven.change_setpoint(setpoint_spinbox.value()))
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout), setpoint_spinbox, setpoint_button
        )

    def create_freeze_widgets(self, layout):
        freeze_button = QPushButton("Freeze for 10s")
        freeze_button.pressed.connect(self.freeze)
        layout_util.add_to_layout(layout_util.add_sublayout(layout, QHBoxLayout), freeze_button)

    def freeze(self):
        for i in range(1, 11):
            print(i)
            time.sleep(1)


class DeveloperMainWindow(MainWindow):
    """
    Add a DeveloperWidget for controlling the oven and use Developer versions of other widgets.
    """

    def __init__(self) -> None:
        MainWindow.__init__(self)
        layout: QHBoxLayout = self.oven_control_tab.layout().itemAt(0).layout()  # type: ignore
        layout.addWidget(DeveloperWidget())


# --------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    INSTRUMENTS.oven = DeveloperOven()
    main(DeveloperMainWindow)



import json
import os
import typing
from pathlib import Path
from typing import Generator, Mapping, Self

from pytest import fixture

from fabrial.utility import serde
from fabrial.utility.serde import TYPE, Deserialize, Json, Serialize

TYPENAME = "tests.utility.serde.serde.MockSerde"
FILES_DIRECTORY = Path(__file__).parent.joinpath("files")


class MockSerde(Deserialize, Serialize):
    """Test class that implements `Deserialize` and `Serialize`."""

    def __init__(self, number: int):
        self.number = number

    @classmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:
        return cls(typing.cast(int, serialized_obj["number"]))

    def serialize(self) -> dict[str, Json]:
        return {"number": self.number}

    def __eq__(self, other: Self):  # type: ignore
        return self.number == other.number


def test_get_type():
    """Tests `serde.get_type()`."""
    assert serde.get_type(TYPENAME) == MockSerde


@fixture
def number() -> int:
    """Fixture to generate the number used by mock objects."""
    return 50


@fixture
def mock_object(number: int) -> MockSerde:
    """Fixture to generate a mock `Serialize`/`Deserialize` object."""
    return MockSerde(number)


def test_deserialize(mock_object: MockSerde, number: int):
    """Tests `serde.deserialize()`."""
    # primitive case
    assert serde.deserialize(47) == 47
    # simple case
    simple_deserializable: dict[str, Json] = {"type": TYPENAME, "number": number}
    assert serde.deserialize(simple_deserializable) == mock_object
    # nested case
    nested_obj: dict[str, Json] = {
        "random_thing": 47,
        "actual_item": {"type": TYPENAME, "number": number},
    }
    assert serde.deserialize(nested_obj) == {"random_thing": 47, "actual_item": mock_object}
    # list case
    list_obj: list[Json] = ["something", 47, {"type": TYPENAME, "number": number}]
    assert serde.deserialize(list_obj) == ["something", 47, mock_object]
    # nested list and dict case
    complex_obj: dict[str, Json] = {
        "random_thing": 47,
        "list": [47, "something", {"type": TYPENAME, "number": number}],
        "nested": {
            "list": [47, "something", {"type": TYPENAME, "number": number}],
            "item": {"type": TYPENAME, "number": number},
        },
    }
    assert serde.deserialize(complex_obj) == {
        "random_thing": 47,
        "list": [47, "something", mock_object],
        "nested": {"list": [47, "something", mock_object], "item": mock_object},
    }


@fixture
def expected_load(mock_object: MockSerde) -> dict[str, int | MockSerde]:
    """Fixture for the expected result of loading from a file."""
    return {"random_thing": 47, "item": mock_object}


def test_load_toml(expected_load: dict[str, int | MockSerde]):
    """Tests `serde.load_toml()`."""
    file = FILES_DIRECTORY.joinpath("load_toml.toml")
    assert serde.load_toml(file) == expected_load


def test_load_json(expected_load: dict[str, int | MockSerde]):
    """Tests `serde.load_json()`."""
    file = FILES_DIRECTORY.joinpath("load_json.json")
    assert serde.load_json(file) == expected_load


@fixture
def save_json_file() -> Generator[Path, None, None]:
    """
    Fixture to provide the path for `serde.test_save_json()` and remove the file after the test
    completes.
    """
    file = FILES_DIRECTORY.joinpath("temp.json")
    yield file
    os.remove(file)  # remove the file when done


def test_save_json(save_json_file: Path):
    """Tests `serde.Serialize.save_json()`."""
    NUMBER = 500
    obj = MockSerde(NUMBER)
    obj.save_json(save_json_file)
    with open(save_json_file, "r") as f:
        assert json.load(f) == {TYPE: TYPENAME, "number": NUMBER}



from os import PathLike
from pathlib import Path

from pytest import fixture

from fabrial.utility.descriptions import FilesDescription, Substitutions, TextDescription


# helper
def compare_lines(actual_text: str, expected_text: str):
    """
    Compare the lines of **actual_text** against the lines of **expected_text**. This
    makes it easier to see where the text doesn't match up.
    """
    for actual_line, expected_line in zip(actual_text.splitlines(), expected_text.splitlines()):
        assert actual_line == expected_line


# helper
def compare_lines_from_file(actual_text: str, expected_file: PathLike[str] | str):
    """Call `compare_lines()` with **actual_text** and the contents of **expected_file**."""
    with open(expected_file, "r") as f:
        expected_text = f.read()
    compare_lines(actual_text, expected_text)


@fixture
def files_directory() -> Path:
    """Provides the location of test files."""
    return Path(__file__).parent.joinpath("files")


# --------------------------------------------------------------------------------------------------
# tests for `FilesDescription` class
def test_everything(files_directory: Path):
    """Tests `descriptions.make_descriptions()` with substitutions in all files."""
    everything_directory = files_directory.joinpath("everything")
    text = FilesDescription(
        everything_directory,
        "everything",
        Substitutions(
            {"SUBSTITUTION": "overview substitution text"},
            {"PARAMETER3": "Parameter3Substitution"},
            {"SUBSTITUTION": "visuals substitution text"},
            {"FILE3": "file3substitution.txt"},
        ),
    ).render()
    compare_lines_from_file(text, everything_directory.joinpath("correct.md"))


def test_substitution_errors(files_directory: Path):
    """Tests `descriptions.make_descriptions()` with missing substitutions."""
    substitution_errors_directory = files_directory.joinpath("substitution_errors")
    # the data directory name doesn't matter here
    # we provide no substitutions when there are substitutions in the files, which should yield the
    # error text
    text = FilesDescription(substitution_errors_directory, "").render()
    compare_lines_from_file(text, substitution_errors_directory.joinpath("correct.md"))


def test_toml_errors(files_directory: Path):
    """Tests `descriptions.make_descriptions()` with TOML errors."""
    toml_errors_directory = files_directory.joinpath("toml_errors")
    # the data directory name doesn't matter here
    text = FilesDescription(toml_errors_directory, "").render()
    compare_lines_from_file(text, toml_errors_directory.joinpath("correct.md"))


def test_missing_files(files_directory: Path):
    """Tests `descriptions.make_descriptions()` with missing description files."""
    missing_files_directory = files_directory.joinpath("missing_files")
    # the data directory name doesn't matter here
    text = FilesDescription(missing_files_directory, "").render()
    compare_lines_from_file(text, missing_files_directory.joinpath("correct.md"))


def test_empty_files(files_directory: Path):
    """Tests `descriptions.make_descriptions()` with all empty files."""
    empty_files_directory = files_directory.joinpath("empty_files")
    text = FilesDescription(empty_files_directory, "Empty Files").render()
    compare_lines_from_file(text, empty_files_directory.joinpath("correct.md"))


# --------------------------------------------------------------------------------------------------
# tests for `TextDescription` class
def test_text(files_directory: Path):
    """Tests `TextDescription.render()`."""
    description_provider = TextDescription(
        "Random Directory",
        "Overview text.",
        {"Parameter1": "Description of Parameter1.", "Parameter2": "Description of Parameter2."},
        {"file1.txt": "Description of file1.txt.", "file2.txt": "Description of file2.txt."},
        "Visuals text.",
    )
    compare_lines_from_file(
        description_provider.render(), files_directory.joinpath("correct_for_text.md")
    )



from PyQt6.QtWidgets import QMenu, QMenuBar

from ..classes import Action


class SequenceMenu(QMenu):
    """Sequence menu."""

    def __init__(self, parent: QMenuBar):
        QMenu.__init__(self, "&Sequence", parent)
        self.cancel = Action(parent, "Cancel Sequence")
        self.cancel.setDisabled(True)  # start disabled
        self.addAction(self.cancel)



from serial.tools import list_ports as ports_util


def list_ports() -> list[str]:
    """Returns a list of available system com ports from pyserial."""
    ports: list[str] = []
    # serial.tools.list_ports.comports() returns a list of ports
    for port_info in ports_util.comports():
        # .device accesses the name of the port, i.e. /dev/ttyS14 on Linux or COM5 on windows
        ports.append(port_info.device)
    return ports



from fabrial import PluginCategory

# notice we import our item
from .random_item import RandomDataItem


def categories() -> list[PluginCategory]:
    # we use "Random" as the name of the category
    # when creating the `RandomDataItem`, we use 5 as the default value
    return [PluginCategory("Random", [RandomDataItem(5)])]



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QGridLayout

from ..custom_widgets import GroupBox, Label
from ..instruments import INSTRUMENTS


class PassiveMonitoringWidget(GroupBox):
    """Widget for monitoring the oven."""

    NULL_TEXT = "-----"

    def __init__(self):
        GroupBox.__init__(self, "Measurements", QGridLayout())
        self.oven = INSTRUMENTS.oven

        self.create_widgets()
        self.connect_signals()

    def create_widgets(self):
        """Create subwidgets."""
        layout: QGridLayout = self.layout()  # type: ignore

        self.temperature_label = Label(self.NULL_TEXT)
        self.setpoint_label = Label(self.NULL_TEXT)

        self.add_widget_row(layout, "Temperature", 0, self.temperature_label)  # oven temperature
        self.add_widget_row(layout, "Setpoint", 1, self.setpoint_label)  # oven setpoint

    def add_widget_row(self, layout: QGridLayout, text: str, row: int, label_widget: Label):
        category_label = Label("Oven " + text + ":")

        font = label_widget.font()
        font.setPointSize(16)
        label_widget.setFont(font)

        layout.addWidget(category_label, row, 0, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(label_widget, row, 1)

    def connect_signals(self):
        self.oven.connectionChanged.connect(self.handle_connection_change)
        self.oven.temperatureChanged.connect(self.handle_temperature_change)
        self.oven.setpointChanged.connect(self.handle_setpoint_change)

    def handle_connection_change(self, connected: bool):
        if not connected:
            self.handle_disconnect()

    def handle_disconnect(self):
        for label in (self.temperature_label, self.setpoint_label):
            label.setText(self.NULL_TEXT)

    def handle_temperature_change(self, temperature: float):
        self.update_label(self.temperature_label, temperature)

    def handle_setpoint_change(self, setpoint: float):
        self.update_label(self.setpoint_label, setpoint)

    def update_label(self, label: Label, value: float):
        label.setText(str(value))



from typing import Iterable

from PyQt6.QtWidgets import QApplication
from pytest import FixtureRequest, fixture

from fabrial.sequence_builder.tree_items import CategoryItem, SequenceItem, TreeItem
from fabrial.utility import sequence_builder

from .mock_item import MockDataItem
from .test_plugins import (
    empty_categories,
    no_entry_point,
    normal,
    same_category1,
    same_category2,
    wrong_entry_point_type,
    wrong_types,
)


def compare_items(actual: Iterable[CategoryItem], expected: Iterable[CategoryItem]):
    # helper to compare just the items and their parents, not the subitems
    def compare(actual: TreeItem, expected: TreeItem):
        # check the parents
        actual_parent = actual.parent()
        expected_parent = expected.parent()
        if actual_parent is None or expected_parent is None:
            assert actual_parent is None and expected_parent is None
        else:
            compare(actual_parent, expected_parent)

        # if one is a `CategoryItem` they should both be `CategoryItem`s
        if isinstance(actual, CategoryItem) or isinstance(expected, CategoryItem):
            assert type(actual) is type(expected)
            assert actual.display_name() == expected.display_name()
        else:  # otherwise they most both be `SequenceItem`s
            assert isinstance(actual, SequenceItem) and isinstance(expected, SequenceItem)
            # I don't love checking a "private" attribute here
            # TODO: find a better solution
            assert actual.item == expected.item

    # helper to compare the items, their parents, and their subitems
    def recursive_compare(actual: TreeItem, expected: TreeItem):
        compare(actual, expected)
        # check the counts
        actual_count = actual.subitem_count()
        assert actual_count == expected.subitem_count()
        for i in range(actual_count):
            sub_actual = actual.get_subitem(i)
            sub_expected = expected.get_subitem(i)
            # this line assumes that all subitems within the range should be valid
            assert sub_actual is not None and sub_expected is not None
            compare(sub_actual, sub_expected)
            recursive_compare(sub_actual, sub_expected)

    for actual_item, expected_item in zip(actual, expected, strict=True):
        recursive_compare(actual_item, expected_item)


@fixture
def expected(request: FixtureRequest):
    """Provides the expected result on a per-test basis."""
    normal = [
        CategoryItem(
            None,
            "Normal1",
            [
                SequenceItem(None, MockDataItem("Normal11", 11)),
                SequenceItem(None, MockDataItem("Normal12", 12)),
            ],
        ),
        CategoryItem(
            None,
            "Normal2",
            [
                SequenceItem(None, MockDataItem("Normal21", 21)),
                SequenceItem(None, MockDataItem("Normal22", 22)),
            ],
        ),
    ]

    wrong_types = [
        CategoryItem(
            None,
            "Right Types",
            [
                SequenceItem(None, MockDataItem("RightTypes1", 1)),
                SequenceItem(None, MockDataItem("RightTypes2", 2)),
            ],
        ),
    ]

    same_category = [
        CategoryItem(
            None,
            "Outer Category",
            [
                CategoryItem(
                    None,
                    "Nested Category",
                    [
                        SequenceItem(None, MockDataItem("Category1Item1", 11)),
                        SequenceItem(None, MockDataItem("Category1Item2", 12)),
                        SequenceItem(None, MockDataItem("Category2Item1", 21)),
                        SequenceItem(None, MockDataItem("Category2Item2", 22)),
                    ],
                ),
                SequenceItem(None, MockDataItem("Category1Outer", 1)),
                SequenceItem(None, MockDataItem("Category2Outer", 2)),
            ],
        )
    ]

    everything = normal + same_category + wrong_types

    function = request.function
    if function is test_normal:
        return normal
    elif function is test_wrong_types:
        return wrong_types
    elif function is test_same_category:
        return same_category
    elif function is test_everything:
        return everything
    raise ValueError("A test used the `expected()` fixture but is not registered for that fixture")


# --------------------------------------------------------------------------------------------------
# tests start here
def test_normal(qapp: QApplication, expected: Iterable[CategoryItem]):
    """Tests a normal, functioning plugin."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {normal.__name__: normal}
    )
    assert len(failure_plugins) == 0 and len(failure_categories) == 0  # no failures expected
    compare_items(items, expected)


def test_empty():
    """Tests a plugin that returns an empty list of categories."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {empty_categories.__name__: empty_categories}
    )
    assert len(failure_plugins) == 0 and len(failure_categories) == 0  # no failures expected
    assert len(items) == 0  # should just be an empty list


def test_no_entry_point():
    """Tests a plugin that is missing the `categories()` entry point."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {no_entry_point.__name__: no_entry_point}
    )
    assert len(items) == 0 and len(failure_plugins) == 1 and len(failure_categories) == 0
    assert no_entry_point.__name__ in failure_plugins


def test_wrong_entry_point_type():
    """Tests a plugin that returns the wrong type from its entry point."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {wrong_entry_point_type.__name__: wrong_entry_point_type}
    )
    assert len(items) == 0 and len(failure_plugins) == 1 and len(failure_categories) == 0
    assert wrong_entry_point_type.__name__ in failure_plugins


def test_wrong_types(qapp: QApplication, expected: Iterable[CategoryItem]):
    """Tests a plugin that returns the correct entry point type, but an item has the wrong type."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {wrong_types.__name__: wrong_types}
    )
    assert len(failure_plugins) == 0 and len(failure_categories) == 1
    assert "Wrong Types" in failure_categories
    compare_items(items, expected)


def test_same_category(qapp: QApplication, expected: Iterable[CategoryItem]):
    """
    Tests a plugin with nested categories that share names (so they should all be grouped together).
    """
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {same_category1.__name__: same_category1, same_category2.__name__: same_category2}
    )
    assert len(failure_plugins) == 0 and len(failure_categories) == 0
    compare_items(items, expected)


def test_everything(qapp: QApplication, expected: Iterable[CategoryItem]):
    """Tests EVERYTHING! (aka tests the combination of all prior modules)."""
    items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
        {
            module.__name__: module
            for module in [
                # arbitrary order
                empty_categories,
                no_entry_point,
                normal,
                same_category1,
                same_category2,
                wrong_entry_point_type,
                wrong_types,
            ]
        }
    )

    assert len(failure_plugins) == 2 and len(failure_categories) == 1
    assert (
        no_entry_point.__name__ in failure_plugins
        and wrong_entry_point_type.__name__ in failure_plugins
    )
    assert "Wrong Types" in failure_categories
    compare_items(items, expected)



from __future__ import annotations

import logging
import os
import typing
from pathlib import Path
from queue import Queue
from typing import TYPE_CHECKING, Callable

from PyQt6.QtCore import QModelIndex
from PyQt6.QtWidgets import QMessageBox, QPushButton

from ..constants.paths.settings import sequence as sequence_paths
from ..custom_widgets import DontShowAgainDialog, YesNoDialog
from ..enums import SequenceCommand
from ..utility import errors
from .exceptions import PluginError
from .lock import DataLock
from .sequence_step import SequenceStep
from .sequence_thread import SequenceThread

if TYPE_CHECKING:
    from ..sequence_builder import SequenceModel
    from ..tabs import SequenceBuilderTab, SequenceDisplayTab


class ValueButton(QPushButton):
    """A button with an associated value."""

    def __init__(self, text: str, value: int):
        QPushButton.__init__(self, text)
        self.value = value


class SequenceRunner:
    """Initialize and run a sequence."""

    def __init__(self):
        self.command_queue: Queue[SequenceCommand] = Queue()
        self.thread: SequenceThread | None = None  # we have to keep a reference to the thread

    def run_sequence(
        self, sequence_tab: SequenceBuilderTab, model: SequenceModel, data_directory: Path
    ):
        """
        Run the sequence. Creates the data directory if it doesn't exist. Logs errors. Returns
        whether the sequence was successfully started.
        """
        if not model.root().subitem_count() > 0:  # do nothing if there are no items
            return False
        # make the data directory
        if not self.create_files(model, data_directory):
            return False
        # create the `SequenceStep`s
        try:
            sequence_steps, step_item_map = self.create_sequence_steps(model)
        except PluginError:
            logging.getLogger(__name__).exception(
                "Item from plugin generated error during sequence construction"
            )
            errors.show_error(
                "Sequence Error",
                "Unable to generate sequence steps, likely due to a faulty plugin. "
                "See the error log for details.",
            )
            return False
        # create the thread
        self.thread = SequenceThread(sequence_steps, data_directory, self.command_queue)
        # connect signals so the application responds to changes in the sequence
        self.connect_signals(self.thread, sequence_tab, model, step_item_map)
        # start
        self.thread.start()
        return True

    def create_files(self, model: SequenceModel, data_directory: Path) -> bool:
        """Create the sequence's root data directory and generate a sequence autosave."""
        # make the directory (does nothing if the directory exists)
        os.makedirs(data_directory, exist_ok=True)
        # check if the directory is empty
        try:
            empty = len(list(os.scandir(data_directory))) == 0
        except OSError:
            logging.getLogger(__name__).exception("Failed to check if data directory is empty")
            errors.show_error(
                "Sequence Error",
                "Unable to start the sequence because of an operating system error. "
                "See the error log for details.",
            )
            return False
        if not empty:
            # ask the user if they are okay with writing to a non-empty directory
            if not DontShowAgainDialog(
                "Note",
                "Data directory is not empty. Proceed?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel,
                QMessageBox.StandardButton.Yes,
                sequence_paths.NON_EMPTY_DIRECTORY_WARNING_FILE,
            ).run():
                return False
        # try to generate an autosave of the sequence
        if not model.to_json(data_directory.joinpath("autosave.json")):
            return YesNoDialog(
                "Minor Error",
                "Failed to generate sequence autosave. "
                "See the error log for details. Continue sequence?",
            ).run()
        return True

    def create_sequence_steps(
        self, model: SequenceModel
    ) -> tuple[list[SequenceStep], dict[int, QModelIndex]]:
        """
        Create `SequenceStep`s from the items in **model**. Logs errors.

        Returns
        -------
        A tuple of (the created steps, a mapping of memory addresses of sequence steps to the
        `QModelIndex` of the corresponding model item).

        Raises
        ------
        ValueError
            The model returned `None` for a subitem. This indicates an error with the codebase and
            is fatal.
        PluginError
            There was an error while creating a `SequenceStep`, which indicates a faulty plugin. The
            sequence cannot start.
        """
        step_item_map: dict[int, QModelIndex] = {}

        # helper function to create all substeps of an item by index
        def create_substeps(index: QModelIndex) -> list[SequenceStep]:
            sequence_steps: list[SequenceStep] = []
            for i in range(model.rowCount(index)):
                subindex = model.index(i, 0, index)
                substeps = create_substeps(subindex)
                subitem = model.get_item(subindex)
                if subitem is None:  # this should never happen
                    raise ValueError(
                        "A subitem was `None` when it should have been a `SequenceItem`"
                    )
                try:
                    sequence_step = subitem.create_sequence_step(substeps)
                except Exception:
                    raise PluginError(f"Error while creating sequence step from item {subitem!r}")
                # we can use the step's ID because steps are not added to/removed from the sequence
                step_item_map[id(sequence_step)] = subindex
                sequence_steps.append(sequence_step)
            return sequence_steps

        return (create_substeps(QModelIndex()), step_item_map)

    def connect_signals(
        self,
        sequence_thread: SequenceThread,
        sequence_tab: SequenceBuilderTab,
        model: SequenceModel,
        step_item_map: dict[int, QModelIndex],
    ):
        """Connect signals at construction."""
        # send the status to the sequence tab
        sequence_thread.statusChanged.connect(sequence_tab.handle_sequence_status_change)
        # make the corresponding step bold/not bold in the model
        sequence_thread.stepStateChanged.connect(
            lambda step_address, running: model.set_bold(step_item_map[step_address], running)
        )
        # show the prompt and send the response
        sequence_thread.promptRequested.connect(self.show_prompt)
        # run the command on the visuals tab
        sequence_thread.plotCommandRequested.connect(
            lambda command: self.run_plot_command(command, sequence_tab.visuals_tab)
        )
        # notify finish
        sequence_thread.finished.connect(lambda: sequence_tab.handle_sequence_state_change(False))

    def show_prompt(
        self, title: str, message: str, options: dict[int, str], receiver: DataLock[int | None]
    ):
        """Show a prompt to the user, then send the response back to the sequence."""
        # create the prompt
        prompt = QMessageBox()
        prompt.setWindowTitle(title)
        prompt.setText(message)
        for value, button_text in options.items():  # add the options from **options**
            prompt.addButton(ValueButton(button_text, value), QMessageBox.ButtonRole.NoRole)
        prompt.exec()  # run the prompt
        selected_value = typing.cast(ValueButton, prompt.clickedButton()).value  # get the result
        receiver.set(selected_value)  # send to the receiver

    def run_plot_command(
        self, command: Callable[[SequenceDisplayTab], None], visuals_tab: SequenceDisplayTab
    ):
        """Run a plot **command**."""
        try:
            command(visuals_tab)
        except Exception:
            logging.getLogger(__name__).exception(
                "Error while running a plot command. This usually indicates that a plugin used an "
                "invalid `PlotHandle` or `LineHandle`"
            )
            self.command_queue.put_nowait(SequenceCommand.RaiseFatal)

    def pause(self):
        """Pause the sequence."""
        self.command_queue.put_nowait(SequenceCommand.Pause)

    def unpause(self):
        """Unpause the sequence."""
        self.command_queue.put_nowait(SequenceCommand.Unpause)

    def cancel(self):
        """Cancel the sequence."""
        self.command_queue.put_nowait(SequenceCommand.Cancel)



from typing import Any, Callable

from PyQt6.QtCore import QObject

from ..classes import Timer


def new_timer(parent: QObject | None, interval_ms: int, *slots: Callable[[], Any]) -> Timer:
    """
    Instantiate a new timer. The **slots** are called immediately.

    Parameters
    ----------
    parent
        The QObject that owns this timer.
    interval_ms
        The timeout in milliseconds.
    slots
        The function(s) to connect the timer's timeout signal to.

    Returns
    -------
    The created timer. The timer is already started. Note that you must keep a reference to the
    timer so it does not get deleted.
    """
    timer = Timer(parent, interval_ms, *slots)
    timer.start_fast()
    return timer



from typing import Any, Self

from PyQt6.QtGui import QShowEvent
from PyQt6.QtWidgets import QCheckBox, QFormLayout, QHBoxLayout, QRadioButton, QVBoxLayout

from .....classes import DescriptionInfo
from .....custom_widgets import Container, DoubleSpinBox, SpinBox
from .....gamry_integration import GAMRY
from .....utility import layout as layout_util
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import EISProcess


class EISWidget(AbstractBaseWidget):
    """Contains entries for EIS experiments."""

    def __init__(self) -> None:
        """Create a new instance."""
        AbstractBaseWidget.__init__(
            self,
            QFormLayout(),
            "Electrochemical Impedance Spectroscopy",
            EISProcess,
            "battery-charge.png",
            DescriptionInfo(
                "electrochemistry",
                "EIS",
                EISProcess.directory_name(),
                DescriptionInfo.Substitutions(
                    parameters_dict={
                        "POTENTIOSTATS": encoding.SELECTED_PSTATS,
                        "DC_VOLTAGE": encoding.DC_VOLTAGE,
                        "VS_EREF": encoding.VS_EREF,
                        "VS_EOC": encoding.VS_EOC,
                        "INITIAL_FREQUENCY": encoding.INITIAL_FREQUENCY,
                        "FINAL_FREQUENCY": encoding.FINAL_FREQUENCY,
                        "POINTS_PER_DECADE": encoding.POINTS_PER_DECADE,
                        "AC_VOLTAGE": encoding.AC_VOLTAGE,
                        "ESTIMATED_IMPEDANCE": encoding.ESTIMATED_IMPEDANCE,
                    }
                ),
            ),
        )
        self.pstat_checkboxes: dict[str, QCheckBox] = dict()
        self.impedance_reader_speed_radiobuttons: list[QRadioButton] = []
        self.voltage_reference_radiobuttons: list[QRadioButton] = []

        self.create_widgets()

    def create_widgets(self) -> None:
        """Create subwidgets."""
        layout: QFormLayout = self.parameter_widget().layout()  # type: ignore

        self.devices_layout = QVBoxLayout()
        device_list_container = Container(self.devices_layout)

        self.DC_voltage_spinbox = DoubleSpinBox(4)
        self.initial_frequency_spinbox = DoubleSpinBox(4)
        self.final_frequency_spinbox = DoubleSpinBox(4)
        self.points_per_decade_spinbox = SpinBox()
        self.AC_voltage_spinbox = DoubleSpinBox(4)
        self.estimated_impedance_spinbox = DoubleSpinBox(4)

        impedance_reader_speed_layout = QHBoxLayout()
        impedance_reader_speed_container = Container(impedance_reader_speed_layout)
        for label in encoding.IMPEDANCE_READER_SPEED_DICT.keys():
            radiobutton = QRadioButton(label)
            self.impedance_reader_speed_radiobuttons.append(radiobutton)
            impedance_reader_speed_layout.addWidget(radiobutton)

        dc_voltage_layout = QHBoxLayout()
        dc_voltage_layout.addWidget(self.DC_voltage_spinbox)
        voltage_reference_layout = QHBoxLayout()
        dc_voltage_layout.addLayout(voltage_reference_layout)
        dc_voltage_container = Container(dc_voltage_layout)
        for label in (encoding.VS_EREF, encoding.VS_EOC):
            radiobutton = QRadioButton(label)
            self.voltage_reference_radiobuttons.append(radiobutton)
            voltage_reference_layout.addWidget(radiobutton)

        layout_util.add_to_form_layout(
            layout,
            (encoding.SELECTED_PSTATS, device_list_container),
            (encoding.DC_VOLTAGE, dc_voltage_container),
            (encoding.INITIAL_FREQUENCY, self.initial_frequency_spinbox),
            (encoding.FINAL_FREQUENCY, self.final_frequency_spinbox),
            (encoding.POINTS_PER_DECADE, self.points_per_decade_spinbox),
            (encoding.AC_VOLTAGE, self.AC_voltage_spinbox),
            (encoding.ESTIMATED_IMPEDANCE, self.estimated_impedance_spinbox),
            ("Optimize for", impedance_reader_speed_container),
        )

    def reload_pstat_list(self, selected_pstats: list[str]) -> None:
        """
        Reload the list of available potentiostats.

        Parameters
        ----------
        selected_pstats
            A list of potentiostat identifiers which should be checked.
        """
        identifiers: list[str] = GAMRY.get_pstat_list()
        layout_util.clear_layout(self.devices_layout)
        self.pstat_checkboxes.clear()

        for identifier in identifiers:
            checkbox = QCheckBox(identifier)
            self.pstat_checkboxes.update({identifier: checkbox})
            if identifier in selected_pstats:
                checkbox.setChecked(True)
            self.devices_layout.addWidget(checkbox)

    def selected_pstats(self) -> list[str]:
        """Get a list of the selected potentiostats. The list contains potentiostat identifiers."""
        selected_pstats: list[str] = []
        for identifier, checkbox in self.pstat_checkboxes.items():
            if checkbox.isChecked():
                selected_pstats.append(identifier)
        return selected_pstats

    def showEvent(self, event: QShowEvent | None):  # overridden
        if event is not None:
            # reload the device list when the window gets opened
            self.reload_pstat_list(self.selected_pstats())
        AbstractBaseWidget.showEvent(self, event)

    @staticmethod
    def allowed_to_create() -> bool:  # overridden
        return GAMRY.is_valid()

    @classmethod
    def from_dict(cls: type[Self], data_as_dict: dict[str, Any]) -> Self:
        widget = cls()
        widget.DC_voltage_spinbox.setValue(data_as_dict[encoding.DC_VOLTAGE])
        widget.initial_frequency_spinbox.setValue(data_as_dict[encoding.INITIAL_FREQUENCY])
        widget.final_frequency_spinbox.setValue(data_as_dict[encoding.FINAL_FREQUENCY])
        widget.points_per_decade_spinbox.setValue(data_as_dict[encoding.POINTS_PER_DECADE])
        widget.AC_voltage_spinbox.setValue(data_as_dict[encoding.AC_VOLTAGE])
        widget.estimated_impedance_spinbox.setValue(data_as_dict[encoding.ESTIMATED_IMPEDANCE])

        selected_speed_option = data_as_dict[encoding.IMPEDANCE_READER_SPEED]
        for radiobutton in widget.impedance_reader_speed_radiobuttons:
            if radiobutton.text() == selected_speed_option:
                radiobutton.setChecked(True)
                break

        selected_voltage_reference = data_as_dict[encoding.DC_VOLTAGE_REFERENCE]
        for radiobutton in widget.voltage_reference_radiobuttons:
            if radiobutton.text() == selected_voltage_reference:
                radiobutton.setChecked(True)
                break

        widget.reload_pstat_list(data_as_dict[encoding.SELECTED_PSTATS])

        return widget

    def to_dict(self) -> dict:
        for radiobutton in self.impedance_reader_speed_radiobuttons:
            if radiobutton.isChecked():
                selected_speed_option = radiobutton.text()
                break

        for radiobutton in self.voltage_reference_radiobuttons:
            if radiobutton.isChecked():
                selected_voltage_reference = radiobutton.text()
                break

        data = {
            encoding.DC_VOLTAGE: self.DC_voltage_spinbox.value(),
            encoding.DC_VOLTAGE_REFERENCE: selected_voltage_reference,
            encoding.INITIAL_FREQUENCY: self.initial_frequency_spinbox.value(),
            encoding.FINAL_FREQUENCY: self.final_frequency_spinbox.value(),
            encoding.POINTS_PER_DECADE: self.points_per_decade_spinbox.value(),
            encoding.AC_VOLTAGE: self.AC_voltage_spinbox.value(),
            encoding.ESTIMATED_IMPEDANCE: self.estimated_impedance_spinbox.value(),
            encoding.IMPEDANCE_READER_SPEED: selected_speed_option,
            encoding.SELECTED_PSTATS: self.selected_pstats(),
        }
        return data



import os
from io import TextIOWrapper
from typing import Any

from .....classes import AbstractGraphingProcess, ProcessRunner
from .....constants.paths.process import filenames
from .....instruments import INSTRUMENTS, InstrumentLocker
from .....utility import dataframe, temperature as temperature_util
from . import encoding


class SetTemperatureProcess(AbstractGraphingProcess):
    """
    Set the oven's temperature and record the temperature while waiting for it to stabilize.

    When subclassing, you should override:
    - `title()` - Returns the title used on the graph.
    - `metadata()` - You should probably call `Process`'s `metadata()` method.
    """

    MEASUREMENT_INTERVAL = 5000
    """Measurement interval in milliseconds."""
    MINIMUM_MEASUREMENTS = 2
    """The minimum number of measurements for stability."""
    TOLERANCE = 1.0  # degrees C
    """How far off a temperature can be from the setpoint before it is considered unstable."""
    GRAPH_FILENAME = "graph.png"
    """The name used for the graph."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractGraphingProcess.__init__(self, runner, data, name)
        self.oven = INSTRUMENTS.oven
        self.setpoint = data[encoding.SETPOINT]

        self.file: TextIOWrapper

    @staticmethod
    def directory_name():
        return "Set Temperature"

    def pre_run(self):
        """Pre-run tasks."""
        self.init_scatter_plot(
            0,
            "Temperature Graph",
            self.title(),
            "Time (seconds)",
            "Temperature (°C)",
            "Oven Temperature",
        )
        self.create_file()

    def run(self):
        try:
            with InstrumentLocker(self.oven):
                self.pre_run()
                if self.change_setpoint():
                    while not self.oven.is_stable():
                        temperature = self.oven.read_temp()
                        if temperature is not None:  # we read successfully
                            self.record_temperature(temperature)
                        else:  # connection problem
                            self.error_pause()
                        if not self.wait(self.MEASUREMENT_INTERVAL, self.oven.is_connected):
                            break
        finally:
            self.post_run()

    def post_run(self):
        """Post-run tasks."""
        self.file.close()
        self.graphing_signals().saveFig.emit(0, os.path.join(self.directory(), self.GRAPH_FILENAME))

    def create_file(self):
        """
        Create the data file and its write headers.

        Returns
        -------
        The created file.
        """
        self.file = temperature_util.create_temperature_file(
            os.path.join(self.directory(), filenames.TEMPERATURES)
        )

    def change_setpoint(self) -> bool:
        """
        Change the oven's setpoint, going into an error state on failure. Retry until the setpoint
        is changed. Returns whether the process should continue.
        """
        if not self.oven.change_setpoint(self.setpoint):
            self.error_pause()
            while not self.oven.change_setpoint(self.setpoint):
                if not self.wait(self.MEASUREMENT_INTERVAL, self.oven.is_connected):
                    return False
        return True

    def record_temperature(self, temperature: float):
        """Record a temperature in a file and on the graph."""
        time_since_start = temperature_util.record_temperature_data(
            self.file, self.start_time(), temperature
        )
        self.graphing_signals().addPoint.emit(0, time_since_start, temperature)

    def title(self) -> str:
        """Get the graph title."""
        return f"Set Temperature ({self.setpoint} °C)"

    def metadata(self):  # overridden
        return dataframe.add_to_dataframe(
            AbstractGraphingProcess.metadata(self), {"Selected Setpoint": self.setpoint}
        )



from .gamry import GamrySettingsTab
from .oven import OvenSettingsTab
from .sequence import SequenceSettingsTab
from .settings_description import SettingsDescriptionWidget
from .window import ApplicationSettingsWindow



from PyQt6.QtWidgets import QLayout

from ..parameter_description import ParameterDescriptionWidget


class SettingsDescriptionWidget(ParameterDescriptionWidget):
    """Widget with two tabs: one for settings and one for description text."""

    def __init__(self, parameter_layout: QLayout | None = None):
        ParameterDescriptionWidget.__init__(self, parameter_layout, "Settings")



import asyncio
import logging
import time
from asyncio import CancelledError, TaskGroup
from pathlib import Path
from queue import Empty, Queue
from typing import Iterable

from PyQt6.QtCore import QThread, pyqtSignal

from ..enums import SequenceCommand, SequenceStatus
from .exceptions import FatalSequenceError
from .lock import DataLock
from .sequence_step import SequenceStep
from .step_runner import StepRunner


class SequenceThread(QThread):
    """A `QThread` where the sequence runs."""

    statusChanged = pyqtSignal(SequenceStatus)
    """
    Emitted when the sequence's status changes.

    Sends
    - The status as a `SequenceStatus`.
    """
    promptRequested = pyqtSignal(str, str, dict, DataLock)  # see `StepRunner`
    plotCommandRequested = pyqtSignal(object)  # see `StepRunner`
    stepStateChanged = pyqtSignal("qint64", bool)  # see `StepRunner`

    def __init__(
        self,
        steps: Iterable[SequenceStep],
        data_directory: Path,
        command_queue: Queue[SequenceCommand],
    ):
        QThread.__init__(self)
        self.steps = steps
        self.data_directory = data_directory
        self.command_queue = command_queue
        # create the runner
        self.runner = StepRunner()
        self.runner.moveToThread(self)
        self.runner.promptRequested.connect(self.promptRequested)
        self.runner.plotCommandRequested.connect(self.plotCommandRequested)
        self.runner.stepStateChanged.connect(self.stepStateChanged)

    def run(self):  # overridden
        try:
            self.statusChanged.emit(SequenceStatus.Active)
            cancelled = not asyncio.run(self.run_actual())
            self.statusChanged.emit(
                SequenceStatus.Cancelled if cancelled else SequenceStatus.Completed
            )
        except* (FatalSequenceError, Exception):
            logging.getLogger(__name__).exception("Sequence raised a fatal exception")
            self.statusChanged.emit(SequenceStatus.FatalError)

    async def run_actual(self) -> bool:
        """
        The actual run function (required so we can do `asyncio.run()`). Returns whether the
        sequence ended normally. This can raise exceptions.
        """
        async with TaskGroup() as task_group:
            sequence_task = task_group.create_task(
                self.runner.run_steps(self.steps, self.data_directory)
            )
            monitor_task = task_group.create_task(self.monitor())
            # if either task ends, cancel the other
            monitor_task.add_done_callback(lambda _: sequence_task.cancel())
            sequence_task.add_done_callback(lambda _: monitor_task.cancel())
        return not sequence_task.cancelled()

    async def monitor(self) -> None:
        """Monitor for messages and commands."""
        while True:
            self.check_commands()
            await asyncio.sleep(0)

    def check_commands(self):
        """Check for commands and process them."""
        try:
            while True:
                command = self.command_queue.get_nowait()
                match command:
                    case SequenceCommand.Pause:
                        self.pause()
                    case SequenceCommand.Cancel:
                        raise CancelledError
                    case SequenceCommand.Unpause:  # this should never run
                        raise ValueError(
                            "Received `SequenceStep.Unpause`, which should be impossible"
                        )
                    case SequenceCommand.RaiseFatal:
                        raise FatalSequenceError("The main thread requested a fatal error")
        except Empty:
            pass

    def pause(self):
        """Pause the sequence."""
        self.statusChanged.emit(SequenceStatus.Paused)
        while True:
            try:
                command = self.command_queue.get_nowait()
                match command:
                    case SequenceCommand.Cancel:
                        raise CancelledError
                    case SequenceCommand.Unpause:
                        self.statusChanged.emit(SequenceStatus.Active)
                        return
                    case SequenceCommand.Pause:  # ignore additional pauses
                        pass
            except Empty:
                pass
            time.sleep(0)  # this is not `asyncio.sleep()`



from __future__ import annotations

import json
import logging
import typing
from os import PathLike
from typing import Iterable, Mapping, Self, Sequence

from PyQt6.QtCore import QModelIndex, Qt
from PyQt6.QtWidgets import QAbstractItemView, QTreeView

from ...classes import Shortcut
from ...utility.serde import Json
from ..tree_models.tree_model import TreeModel

EXPANDED = "expanded"
SUBITEMS = "subitems"
# TODO: move SUBITEMS to a constant


class TreeView[Model: TreeModel](QTreeView):  # type: ignore
    """Custom TreeView with support for copy, cut, paste, and delete (and drag and drop)."""

    def __init__(self, model: Model):
        QTreeView.__init__(self)

        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setExpandsOnDoubleClick(False)
        self.setModel(model)
        self.connect_signals()
        self.create_shortcuts()

    def create_shortcuts(self) -> Self:
        """Create shortcuts at construction."""
        Shortcut(
            self, "Ctrl+C", self.copy_event, context=Qt.ShortcutContext.WidgetWithChildrenShortcut
        )
        return self

    def items_editable(self) -> bool:
        """Whether the items are editable."""
        return False

    def init_view_state_from_json(self, file: PathLike[str] | str) -> bool:
        """
        Recursively set the view state based on the contents of **file**. Returns whether the
        operation succeeded. Logs errors.
        """
        model = self.model()

        # initialize the view states of all subitems of **index**
        def recursively_init_state(index: QModelIndex, view_states: Sequence[Mapping[str, Json]]):
            for i, subitem_view_state in enumerate(view_states):
                subitem_index = model.index(i, 0, index)
                if subitem_view_state[EXPANDED]:
                    self.expand(subitem_index)
                recursively_init_state(
                    subitem_index,
                    typing.cast(Sequence[Mapping[str, Json]], subitem_view_state[SUBITEMS]),
                )

        try:
            with open(file) as f:
                view_states: Sequence[Mapping[str, Json]] = json.load(f)
            recursively_init_state(self.rootIndex(), view_states)
            return True
        except Exception:
            logging.getLogger(__name__).info(
                "Failed to initialize view state from file", exc_info=True
            )
            self.expandAll()  # just expand everything if we can't load from a file
            return False

    def save_view_state_to_json(self, file: PathLike[str] | str):
        """
        Save the view state to a JSON **file**. Returns whether the operation succeeded. Logs
        errors.
        """
        model = self.model()

        # gets the view state of all subitems
        def get_state(index: QModelIndex) -> list[dict[str, Json]]:
            subitem_states: list[dict[str, Json]] = []
            for i in range(model.rowCount(index)):
                subitem_index = model.index(i, 0, index)
                subitem_states.append(
                    {
                        EXPANDED: self.isExpanded(subitem_index),
                        SUBITEMS: get_state(subitem_index),
                    }
                )
            return subitem_states

        state = get_state(self.rootIndex())
        try:
            with open(file, "w") as f:
                json.dump(state, f)
            return True
        except Exception:
            logging.getLogger(__name__).exception("Failed to save view state to file")
            return False

    def connect_signals(self) -> Self:
        """Connect signals at construction."""
        self.expanded.connect(lambda index: self.model().expand_event(index))
        self.collapsed.connect(lambda index: self.model().collapse_event(index))
        self.doubleClicked.connect(lambda index: self.open_event([index]))
        return self

    def model(self) -> Model:  # overridden for typing
        """Get this view's associated model."""
        return typing.cast(Model, QTreeView.model(self))

    def copy_event(self):
        """Copy items to the clipboard."""
        self.model().copy_items(self.selectedIndexes())

    def open_event(self, indexes: Iterable[QModelIndex]):
        """Perform the items' "open" functions."""
        for index in indexes:
            item = self.model().get_item(index)
            if item is not None:
                if item.open_event(self.items_editable()):
                    self.setExpanded(index, not self.isExpanded(index))



from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout

from .....classes import DescriptionInfo
from .....custom_widgets import TemperatureSpinBox
from .....instruments import INSTRUMENTS
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import SetNoStabilizeProcess


class SetNoStabilizeWidget(AbstractBaseWidget):
    """Set the setpoint without stabilizing."""

    DISPLAY_NAME_PREFIX = "Set Oven Temperature (No Stabilization)"

    def __init__(self):
        layout = QFormLayout()
        AbstractBaseWidget.__init__(
            self,
            layout,
            self.DISPLAY_NAME_PREFIX,
            SetNoStabilizeProcess,
            "thermometer.png",
            DescriptionInfo(
                "temperature",
                "set_no_stabilize",
                SetNoStabilizeProcess.directory_name(),
                DescriptionInfo.Substitutions(parameters_dict={"SETPOINT": encoding.SETPOINT}),
            ),
        )

        self.temperature_spinbox = TemperatureSpinBox(INSTRUMENTS.oven)
        self.temperature_spinbox.textChanged.connect(
            lambda value_as_str: self.setWindowTitle(
                f"{self.DISPLAY_NAME_PREFIX} ({value_as_str} degrees)"
            )
        )
        layout.addRow(encoding.SETPOINT, self.temperature_spinbox)

    @classmethod
    def from_dict(cls, data_as_dict: dict[str, Any]) -> Self:
        widget = cls()
        widget.temperature_spinbox.setValue(data_as_dict[encoding.SETPOINT])
        return widget

    def to_dict(self) -> dict[str, Any]:
        return {encoding.SETPOINT: self.temperature_spinbox.value()}



from typing import Iterable

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QTabWidget, QWidget


class TabWidget(QTabWidget):
    """
    `QTabWidget` that creates the tabs in the constructor, looks better, and auto-focuses a tab
    when it is selected.

    Parameters
    ----------
    tabs
        Tuples of (widget to use for tab, tab name, tab icon).
    """

    def __init__(self, tabs: Iterable[tuple[QWidget, str, QIcon]]):
        QTabWidget.__init__(self)
        self.setDocumentMode(True)

        for tab, name, icon in tabs:
            self.addTab(tab, icon, name)

        self.currentChanged.connect(lambda index: self.widget(index).setFocus())  # type: ignore



from PyQt6.QtWidgets import QProgressBar


class ProgressBar(QProgressBar):
    """
    `QProgressBar` with easier methods.

    Parameters
    ----------
    minimum
        The minimum value of the progressbar.
    maximum
        The maximum value of the progressbar.
    increment_value
        How much increment() increments the value by. Defaults to 1.
    """

    def __init__(self, minimum: int, maximum: int, increment_value: int = 1):
        QProgressBar.__init__(self)
        self.increment_value = increment_value
        self.setMinimum(minimum)
        self.setMaximum(maximum)

    def increment(self):
        """Increment the progressbar."""
        self.setValue(self.value() + self.increment_value)



from typing import Sequence

from PyQt6.QtGui import QIcon

from ...utility import images
from . import tree_item
from .tree_item import TreeItem


class CategoryItem(TreeItem):
    """
    An item representing a category. Categories do nothing by themselves; they just hold other
    items.

    Parameters
    ----------
    parent
        The item's parent.
    category_name
        The name displayed on the item.
    subitems
        The item's subitems. The items' parent is automatically set to this item.
    """

    def __init__(self, parent: TreeItem | None, category_name: str, subitems: Sequence[TreeItem]):
        self.parent_item = parent
        self.name = category_name
        # these two need to be instance attributes instead of class attributes because you have to
        # construct a `QApplication` before a `QIcon`
        self.collapsed_icon = images.make_icon("folder-horizontal.png")
        self.expanded_icon = images.make_icon("folder-horizontal-open.png")

        self.item_icon = self.collapsed_icon
        for item in subitems:
            item.set_parent(self)
        self.subitems = subitems

    def serialize(self):  # implementation
        raise NotImplementedError("`CategoryItem`s do not support serialization")

    def parent(self) -> TreeItem | None:  # implementation
        return self.parent_item

    def set_parent(self, parent: TreeItem | None):  # implementation
        self.parent_item = parent

    def display_name(self) -> str:  # overridden
        return self.name

    def icon(self) -> QIcon:  # overridden
        return self.item_icon

    def subitem_count(self) -> int:  # implementation
        return len(self.subitems)

    def index(self, item: TreeItem) -> int | None:  # implementation
        return tree_item.index(self.subitems, item)

    def get_subitem(self, index: int) -> TreeItem | None:  # implementation
        return tree_item.get_subitem(self.subitems, index)

    def expand_event(self):  # overridden
        self.item_icon = self.expanded_icon

    def collapse_event(self):  # overridden
        self.item_icon = self.collapsed_icon

    # debugging
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} {{ "
            f"display_name: {self.display_name()!r}, "
            f"subitems: {self.subitems!r} }}"
        )



import json

from PyQt6.QtWidgets import QFormLayout

from ...constants import APP_NAME
from ...constants.paths.settings import oven as oven_settings
from ..augmented import DoubleSpinBox, SpinBox
from .settings_description import SettingsDescriptionWidget


class OvenSettingsTab(SettingsDescriptionWidget):
    """Oven settings."""

    MINIMUM_MEASUREMENT_INTERVAL = 50

    def __init__(self) -> None:
        try:
            with open(oven_settings.SAVED_SETTINGS_FILE, "r") as f:
                saved_settings = json.load(f)
        except Exception:
            with open(oven_settings.DEFAULT_SETTINGS_FILE, "r") as f:
                saved_settings = json.load(f)
        number_of_decimals = saved_settings[oven_settings.NUM_DECIMALS]

        layout = QFormLayout()
        SettingsDescriptionWidget.__init__(self, layout)
        self.set_description_from_file(
            oven_settings.DESCRIPTION_FOLDER,
            oven_settings.DESCRIPTION_FILENAME,
            {
                "APPLICATION_NAME": APP_NAME,
                "MINIMUM_TEMPERATURE": oven_settings.MIN_TEMPERATURE,
                "MAXIMUM_TEMPERATURE": oven_settings.MAX_TEMPERATURE,
                "MEASUREMENT_INTERVAL": oven_settings.MEASUREMENT_INTERVAL,
                "STABILITY_TOLERANCE": oven_settings.STABILITY_TOLERANCE,
                "MINIMUM_STABILITY_MEASUREMENTS": oven_settings.MINIMUM_STABILITY_MEASUREMENTS,
                "STABILITY_CHECK_INTERVAL": oven_settings.STABILITY_MEASUREMENT_INTERVAL,
                "TEMPERATURE_REGISTER": oven_settings.TEMPERATURE_REGISTER,
                "SETPOINT_REGISTER": oven_settings.SETPOINT_REGISTER,
                "NUMBER_OF_DECIMALS": oven_settings.NUM_DECIMALS,
            },
        )

        self.minimum_temperature_spinbox = DoubleSpinBox(
            number_of_decimals,
            -DoubleSpinBox.LARGEST_FLOAT,
            initial_value=saved_settings[oven_settings.MIN_TEMPERATURE],
        )
        self.maximum_temperature_spinbox = DoubleSpinBox(
            number_of_decimals,
            -DoubleSpinBox.LARGEST_FLOAT,
            initial_value=saved_settings[oven_settings.MAX_TEMPERATURE],
        )
        self.measurement_interval_spinbox = SpinBox(
            self.MINIMUM_MEASUREMENT_INTERVAL,
            initial_value=saved_settings[oven_settings.MEASUREMENT_INTERVAL],
        )
        self.stability_tolerance_spinbox = DoubleSpinBox(
            number_of_decimals, initial_value=saved_settings[oven_settings.STABILITY_TOLERANCE]
        )
        self.minimum_stability_measurements_spinbox = SpinBox(
            initial_value=saved_settings[oven_settings.MINIMUM_STABILITY_MEASUREMENTS]
        )
        self.stability_measurement_interval_spinbox = SpinBox(
            self.MINIMUM_MEASUREMENT_INTERVAL,
            initial_value=saved_settings[oven_settings.STABILITY_MEASUREMENT_INTERVAL],
        )
        self.temperature_register_spinbox = SpinBox(
            initial_value=saved_settings[oven_settings.TEMPERATURE_REGISTER]
        )
        self.setpoint_register_spinbox = SpinBox(
            initial_value=saved_settings[oven_settings.SETPOINT_REGISTER]
        )
        self.number_of_decimals_spinbox = SpinBox(
            initial_value=saved_settings[oven_settings.NUM_DECIMALS]
        )

        self.spinbox_dict: dict[str, SpinBox | DoubleSpinBox] = {
            oven_settings.MIN_TEMPERATURE: self.minimum_temperature_spinbox,
            oven_settings.MAX_TEMPERATURE: self.maximum_temperature_spinbox,
            oven_settings.MEASUREMENT_INTERVAL: self.measurement_interval_spinbox,
            oven_settings.STABILITY_TOLERANCE: self.stability_tolerance_spinbox,
            oven_settings.MINIMUM_STABILITY_MEASUREMENTS: (
                self.minimum_stability_measurements_spinbox
            ),
            oven_settings.STABILITY_MEASUREMENT_INTERVAL: (
                self.stability_measurement_interval_spinbox
            ),
            oven_settings.TEMPERATURE_REGISTER: self.temperature_register_spinbox,
            oven_settings.SETPOINT_REGISTER: self.setpoint_register_spinbox,
            oven_settings.NUM_DECIMALS: self.number_of_decimals_spinbox,
        }

        for name, spinbox in self.spinbox_dict.items():
            layout.addRow(name, spinbox)

        self.number_of_decimals_spinbox.valueChanged.connect(self.handle_decimal_change)

    def handle_decimal_change(self, decimal_count: int):
        """Handle the decimal number changing."""
        for spinbox in (
            self.maximum_temperature_spinbox,
            self.minimum_temperature_spinbox,
            self.stability_tolerance_spinbox,
        ):
            spinbox.setDecimals(decimal_count)

    def save_on_close(self):
        """Call this when closing the settings window to save settings."""
        settings_dict = {key: spinbox.value() for key, spinbox in self.spinbox_dict.items()}
        with open(oven_settings.SAVED_SETTINGS_FILE, "w") as f:
            json.dump(settings_dict, f)



from typing import Any, Callable

from PyQt6.QtCore import QObject, Qt, pyqtBoundSignal
from PyQt6.QtGui import QAction, QKeySequence, QShortcut


class Action(QAction):
    """
    Easy QAction class.

    Parameters
    ----------
    parent
        The QObject to tie this action to.
    name
        The displayed name of this action in a QMenuBar.
    actions
        The function(s) this action calls.
    status_tip
        Optional text that is shown when a user hovers over the action.
    shortcut
        An optional keyboard shortcut in the form of "Ctrl+A".
    shortcut_context
        The shortcut context. Defaults to `Qt.ShortcutContext.WindowShortcut`.
    """

    def __init__(
        self,
        parent: QObject,
        name: str,
        *actions: Callable[..., Any] | pyqtBoundSignal,
        status_tip: str | None = None,
        shortcut: str | None = None,
        shortcut_context: Qt.ShortcutContext = Qt.ShortcutContext.WindowShortcut,
    ):
        QAction.__init__(self, name, parent)
        for action in actions:
            self.triggered.connect(action)
        self.setStatusTip(status_tip)
        if shortcut is not None:
            self.setShortcut(QKeySequence(shortcut))
            self.setShortcutContext(shortcut_context)


class Shortcut(QShortcut):
    """
    Easy QShortcut class.

    Parameters
    ----------
    parent
        The QObject to tie this shortcut to.
    key
        A keyboard shortcut in the form of "Ctrl+A".
    actions
        The function(s) this shortcut executes.
    context
        The shortcut context. Default `WindowShortcut`.
    """

    def __init__(
        self,
        parent: QObject,
        key: str,
        *actions: Callable[[], Any],
        context: Qt.ShortcutContext = Qt.ShortcutContext.WindowShortcut,
    ):
        QShortcut.__init__(self, QKeySequence(key), parent, context=context)
        for action in actions:
            self.activated.connect(action)



import os
from typing import Any

from .....classes import AbstractBackgroundProcess, ProcessRunner
from .....constants.paths.process import filenames
from .....instruments import INSTRUMENTS
from .....utility import temperature as temperature_util
from . import encoding


class BackgroundTemperatureProcess(AbstractBackgroundProcess):
    """Collect and record temperature samples in the background."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractBackgroundProcess.__init__(self, runner, data, name)
        self.measurement_interval = data[encoding.MEASUREMENT_INTERVAL]

    def run(self):
        file = temperature_util.create_temperature_file(
            os.path.join(self.directory(), filenames.TEMPERATURES)
        )
        oven = INSTRUMENTS.oven
        start_time = self.start_time()

        while True:
            temperature = oven.read_temp()
            temperature_util.record_temperature_data(file, start_time, temperature)
            if not self.wait(self.measurement_interval):
                break

        file.close()

    @staticmethod
    def directory_name():
        return "Background Temperature Monitoring"



import time

FORMAT_SPECIFIER = "%d %B %Y %I:%M:%S %p"


def get_datetime(time_since_epoch: float | None = None) -> str:
    """
    Converts a float (as returned by **time.time()**) to a string in
    Day Month Year Hours:Minutes:Seconds AM/PM format. If **time_since_epoch** is **None**, use the
    current time instead.
    """
    return time.strftime(FORMAT_SPECIFIER, time.localtime(time_since_epoch))


def get_file_friendly_datatime(time_since_epoch: float | None = None) -> str:
    """
    Calls `get_datetime()` and replaces colons (:) with a unicode character that looks like a colon.
    """
    return get_datetime(time_since_epoch).replace(":", "ː")



from PyQt6.QtCore import QSize, Qt
from PyQt6.QtWidgets import QFrame, QTextBrowser


class MarkdownView(QTextBrowser):
    """Label with support for markdown."""

    def __init__(self):
        QTextBrowser.__init__(self)
        self.setReadOnly(True)
        self.setTextInteractionFlags(
            self.textInteractionFlags() | Qt.TextInteractionFlag.LinksAccessibleByMouse
        )
        self.setOpenExternalLinks(True)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setViewportMargins(20, 20, 20, 20)  # this just looks better

    def sizeHint(self) -> QSize:
        return QTextBrowser.sizeHint(self) * 2  # looks better



from .actions import Action, Shortcut
from .exceptions import FatalSequenceError, PluginError
from .lock import DataLock
from .metaclasses import QABC, ABCQObjectMeta, QABCMeta, QProtocol, QProtocolMeta
from .plotting import (
    LineData,
    LineHandle,
    LineIndex,
    LineSettings,
    PlotHandle,
    PlotIndex,
    PlotSettings,
)
from .sequence_runner import SequenceRunner
from .sequence_step import SequenceStep
from .sequence_thread import SequenceThread
from .step_runner import StepRunner
from .timer import Timer



from PyQt6.QtWidgets import QPushButton, QVBoxLayout

from ..custom_widgets import GroupBox, Label, OkDialog, TemperatureSpinBox
from ..instruments import INSTRUMENTS
from ..utility import layout as layout_util


class SetpointWidget(GroupBox):
    """Widget for changing the setpoint."""

    def __init__(self):
        GroupBox.__init__(self, "Setpoint", QVBoxLayout())
        self.oven = INSTRUMENTS.oven

        self.create_widgets()
        self.connect_widgets()
        self.connect_signals()

    def create_widgets(self):
        """Create subwidgets."""
        layout: QVBoxLayout = self.layout()  # type: ignore

        self.setpoint_spinbox = TemperatureSpinBox(self.oven)
        self.button = QPushButton("Change Setpoint")
        layout_util.add_to_layout(layout, Label("Setpoint"), self.setpoint_spinbox, self.button)

    def connect_widgets(self):
        """Connect internal widget signals."""
        self.button.pressed.connect(self.change_setpoint)
        # trigger the command when Enter is pressed
        self.setpoint_spinbox.connect_to_button(self.button)

    def connect_signals(self):
        """Connect external signals."""
        # oven connection
        self.oven.connectionChanged.connect(
            lambda connected: self.update_button_states(connected, self.oven.is_unlocked())
        )
        # oven lock
        self.oven.lockChanged.connect(
            lambda unlocked: self.update_button_states(self.oven.is_connected(), unlocked)
        )

    def update_button_states(self, connected: bool, unlocked: bool):
        """Update the states of buttons."""
        # enable the button if the oven is connected and unlocked
        self.button.setEnabled(connected and unlocked)

    def change_setpoint(self):
        """Change the oven's setpoint."""
        # intentionally not locking the oven because this operation is fast
        success = self.oven.change_setpoint(self.setpoint_spinbox.value())
        if not success:
            # this will probably never run, but it could if the read operation fails
            OkDialog("Error", "Failed to change setpoint, please try again.").exec()



from types import ModuleType
from typing import Any, Iterable, Mapping, Self

from PyQt6.QtCore import QModelIndex, QSize, Qt

from ...utility import errors, sequence_builder
from ..tree_items import CategoryItem
from .tree_model import TreeModel


class OptionsModel(TreeModel[CategoryItem]):
    """
    `TreeModel` for the sequence options.

    Parameters
    ----------
    items
        The direct subitems of the root item.
    """

    def __init__(self, items: Iterable[CategoryItem]):
        TreeModel.__init__(self, "Options", items)

    @classmethod
    def from_plugins(cls, plugin_modules: Mapping[str, ModuleType]) -> Self:
        """
        Create the model from the application's available plugins. Logs errors.

        This calls `plugins.items_from_plugins()`.

        Parameters
        ----------
        plugin_modules
            A list of modules that represent plugins for the application.

        """
        items, failure_plugins, failure_categories = sequence_builder.items_from_plugins(
            plugin_modules
        )
        # if anything failed to load, report it to the user
        if len(failure_plugins) > 0 or len(failure_categories) > 0:
            message = ""
            if len(failure_plugins) > 0:
                message += f"Failed to load items from plugins:\n\n{", ".join(failure_plugins)}\n\n"
            if len(failure_categories) > 0:
                message += f"Failed to load categories:\n\n{", ".join(failure_categories)}\n\n"
            message += "See the error log for details."
            errors.show_error_delayed("Plugin Error", message)
        return cls(items)  # return the new instance

    def data(self, index: QModelIndex, role: int | None = None) -> Any:  # implementation
        if not index.isValid():
            return None
        item = self.get_item(index)
        if item is not None:
            match role:
                case Qt.ItemDataRole.DisplayRole:
                    return item.display_name()
                case Qt.ItemDataRole.DecorationRole:
                    return item.icon()
                case Qt.ItemDataRole.SizeHintRole:
                    return QSize(0, 23)
        return None

    def supportedDragActions(self) -> Qt.DropAction:  # overridden
        return Qt.DropAction.CopyAction



from __future__ import annotations

from abc import abstractmethod
from typing import Iterable, MutableSequence, Protocol, Sequence

from PyQt6.QtGui import QIcon

from ...utility.serde import Json


class TreeItem(Protocol):
    """Interface representing items in a tree model."""

    @abstractmethod
    def serialize(self) -> Json:
        """Serialize the object into a JSON-like structure."""
        ...

    @abstractmethod
    def parent(self) -> TreeItem | None:
        """Get the item's parent, which is `None` if this is the root item."""
        ...

    @abstractmethod
    def set_parent(self, parent: TreeItem | None):
        """Set the item's parent."""
        ...

    def display_name(self) -> str:
        """Get the name displayed on the item."""
        return ""

    def icon(self) -> QIcon:
        """Get the icon displayed on the item."""
        return QIcon()

    def supports_subitems(self) -> bool:
        """Whether the item supports subitems."""
        return True

    def supports_dragging(self) -> bool:
        """Whether the item supports being dragged."""
        return False

    @abstractmethod
    def subitem_count(self) -> int:
        """Get the number of subitems."""
        ...

    def has_subitems(self) -> bool:
        """Whether this item contains subitems."""
        return self.subitem_count() > 0

    @abstractmethod
    def index(self, item: TreeItem) -> int | None:
        """
        Try to find the index of **item** in this item's subitems. Returns `None` if **item** is not
        found.
        """
        ...

    def index_in_parent(self) -> int | None:
        """
        Try to find the index if this item in its parent. Returns `None` if this item is not a
        subitem of its parent.
        """
        ...
        parent = self.parent()
        if parent is None:
            return None
        return parent.index(self)

    @abstractmethod
    def get_subitem(self, index: int) -> TreeItem | None:
        """Get the subitem at **index**."""
        ...

    def open_event(self, editable: bool) -> bool:
        """
        Handle the item being "opened" (i.e. double clicked).

        Parameters
        ----------
        editable
            Whether the item should be editable.

        Returns
        -------
        Whether the item's expansion state should be toggled.
        """
        return True

    def expand_event(self):
        """Handle the item expanding."""
        return

    def collapse_event(self):
        """Handle the item collapsing."""
        return


class GenericTreeItem[SubItem: TreeItem](TreeItem):
    """`TreeItem` with a known subitem type."""

    @abstractmethod
    def get_subitem(self, index: int) -> SubItem | None:  # overridden for typing
        ...


class MutableTreeItem[SubItem: TreeItem](GenericTreeItem[SubItem]):
    """`GenericTreeItem` with support for adding/removing subitems."""

    @abstractmethod
    def append_subitems(self, items: Iterable[SubItem]):
        """Append all **items** to this item's list of subitems."""
        ...

    @abstractmethod
    def insert_subitems(self, start: int, items: Iterable[SubItem]):
        """Insert all **items**, starting at **start**, with the newest subitems on top."""
        ...

    @abstractmethod
    def remove_subitems(self, start: int, count: int):
        """Remove **count** subitems starting at **start**."""
        ...


# --------------------------------------------------------------------------------------------------
# helper functions for external use
def index[SubItem: TreeItem](items: Sequence[SubItem], item: SubItem) -> int | None:
    """Helper function for `TreeItem.index()`. Linear searches for **item** by memory address."""
    # we don't use `Sequence.index()` here because if `SubItem` implements `__eq__()` we might
    # find the wrong item. We need to specifically compare by memory address, not value
    for i, subitem in enumerate(items):
        if subitem is item:
            return i
    return None


def get_subitem[SubItem: TreeItem](items: Sequence[SubItem], index: int) -> SubItem | None:
    """Helper function for `TreeItem.get_subitem`."""
    try:
        return items[index]
    except IndexError:
        return None


def append_subitems[SubItem: TreeItem](
    parent: MutableTreeItem[SubItem], items: MutableSequence[SubItem], subitems: Iterable[SubItem]
):
    """Helper function for `TreeItem.append_subitem`. Appends all **items** to **subitems**."""
    for item in subitems:
        item.set_parent(parent)
    items.extend(subitems)


def insert_subitems[SubItem: TreeItem](
    parent: MutableTreeItem[SubItem],
    items: MutableSequence[SubItem],
    start: int,
    subitems: Iterable[SubItem],
):
    """
    Helper function for `TreeItem.insert_subitems()`. Inserts all **items** into **subitems**,
    starting at **start**.
    """
    for i, item in enumerate(subitems):
        item.set_parent(parent)
        items.insert(start + i, item)


def remove_subitems[SubItem: TreeItem](items: MutableSequence[SubItem], start: int, count: int):
    """
    Helper function for `TreeItem.remove_subitems()`. Removes **count** items from **items**
    starting at **start**.
    """
    del items[start : start + count]



from typing import Any

from .....classes import AbstractForegroundProcess, ProcessRunner
from . import encoding


class HoldProcess(AbstractForegroundProcess):
    """Hold for a duration."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractForegroundProcess.__init__(self, runner, data, name)
        hours: int = data[encoding.HOURS]
        minutes: int = data[encoding.MINUTES]
        seconds: int = data[encoding.SECONDS]
        self.duration = (hours * 3600 + minutes * 60 + seconds) * 1000  # in milliseconds

    def run(self):
        self.wait(self.duration)  # I didn't think it would be that simple lmao

    @staticmethod
    def directory_name():
        return "Hold"



from typing import Iterable

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QComboBox

MAX_VISIBLE_ITEMS = 20


class ComboBox(QComboBox):
    """QComboBox that doesn't show all entries at once."""

    # signal to detect when the combobox is pressed
    pressed = pyqtSignal()

    def __init__(self, *args, **kwargs):
        QComboBox.__init__(self, *args, **kwargs)
        self.setStyleSheet("combobox-popup: 0")
        self.setMaxVisibleItems(MAX_VISIBLE_ITEMS)

    def setCurrentIndexSilent(self, index: int):
        """Update the current index without emitting signals."""
        self.blockSignals(True)
        self.setCurrentIndex(index)
        self.blockSignals(False)

    def setCurrentTextSilent(self, text: str | None):
        """Update the current text without emitting signals."""
        self.blockSignals(True)
        self.setCurrentText(text)
        self.blockSignals(False)

    def clearSilent(self):
        """Clear the combobox entries without emitting signals."""
        self.blockSignals(True)
        self.clear()
        self.blockSignals(False)

    def addItemsSilent(self, items: Iterable[str | None]):
        """Add items to the combobox without emitting signals."""
        self.blockSignals(True)
        self.addItems(items)
        self.blockSignals(False)

    # ----------------------------------------------------------------------------------------------
    # overridden methods
    def showPopup(self):
        self.pressed.emit()
        QComboBox.showPopup(self)



import typing
from typing import Literal

import pyqtgraph as pg
from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QApplication, QFileDialog, QHBoxLayout, QVBoxLayout
from pyqtgraph.exporters import ImageExporter

from ..classes import LineData, Shortcut
from ..utility import errors, layout as layout_util
from .augmented import Button, Widget


def get_text_color() -> QColor:
    """Get the application's current text color."""
    return QApplication.palette().windowText().color()


def get_background_color() -> QColor:
    """Get the application's current background color."""
    return QApplication.palette().window().color()


class PlotItem(pg.PlotItem):
    """Plot item that automatically uses the OS color theme. This is not the widget."""

    LABEL_SIZE = "14pt"
    TITLE_SIZE = "20pt"
    DEFAULT_POINT_SIZE = 7

    def __init__(self) -> None:
        """Create a new PlotItem."""
        pg.PlotItem.__init__(self)
        self.lines: list[LineData] = []
        self.init_plot()

    def init_plot(self):
        """Initialize the plot during construction."""
        self.getViewBox().setBackgroundColor(get_background_color())
        self.addLegend()
        typing.cast(pg.LegendItem, self.legend).setLabelTextColor(get_text_color())
        self.recolor_axis("left")
        self.recolor_axis("bottom")

    def recolor_axis(self, axis_name: Literal["left", "right", "bottom", "top"]):
        """Recolor an axis during construction."""
        axis: pg.AxisItem = self.getAxis(axis_name)
        text_color = get_text_color()
        axis.setPen(text_color)
        axis.setTextPen(text_color)

    def set_label(
        self, axis_name: Literal["left", "right", "bottom", "top"], label: str | None, **kwargs
    ):
        """Set the label text for an axis."""
        axis: pg.AxisItem = self.getAxis(axis_name)
        axis.setLabel(label, **{"font-size": self.LABEL_SIZE}, **kwargs)

    def set_title(self, title: str | None, **kwargs):
        """Set the plot's title."""
        self.setTitle(title, size=self.TITLE_SIZE, color=get_text_color(), **kwargs)

    def reset(self):
        """Reset the plot to it's original state."""
        self.clear()
        for axis_name in ("left", "right", "top", "bottom"):
            self.set_label(axis_name, None)
        self.set_title(None)
        self.setLogMode(False, False)

    def line_count(self) -> int:
        """Get the number of lines on this item."""
        return len(self.lines)

    def plot(
        self,
        x_data: list[float],
        y_data: list[float],
        legend_label: str,
        line_color: str | None,
        line_width: float | None,
        symbol: str,
        symbol_color: str,
        symbol_size: int,
    ) -> LineData:
        """
        Plot a new line on top of the current lines. Stores the plotted item internally.

        Parameters
        ----------
        x_data
            The x-data.
        y_data
            The y-data.
        legend_label
            The label for this line in the legend.
        line_color
            The color of the line. If set to `None`, there will be no line.
        line_width
            The width of the line. If set to `None`, there will be no line.
        symbol
            The symbol to use, i.e. "o" for a dot.
        symbol_size
            The point size.
        symbol_color
            The color of the points.

        Returns
        -------
        A reference to the plotted line.
        """
        if line_color is None or line_width is None:
            line_pen = None
        else:
            line_pen = pg.mkPen(line_color, width=line_width)
        line = pg.PlotItem.plot(
            self,
            x_data,
            y_data,
            name=legend_label,
            pen=line_pen,
            symbol=symbol,
            symbolSize=symbol_size,
            symbolBrush=symbol_color,
            symbolPen=pg.mkPen(color=symbol_color),
        )
        line_data = LineData(line, x_data, y_data)
        self.lines.append(line_data)
        return line_data

    def add_point(self, x: float, y: float, line_index: int):
        """
        Add a point to the line at **line_index**.

        Raises
        ------
        IndexError
            **line_index** is out of range.
        """
        self.lines[line_index].add_point(x, y)

    def export_to_image(self, file: str):
        """Export the item to an image. This uses the item's current dimensions."""
        exporter = ImageExporter(self)
        exporter.export(file)


class PlotView(pg.PlotWidget):
    """Container for a PlotItem. Capable of exporting itself as an image."""

    def __init__(self):
        """Create a new PlotView."""
        self.plot_item = PlotItem()
        pg.PlotWidget.__init__(self, background=get_background_color(), plotItem=self.plot_item)


class PlotWidget(Widget):
    """Contains a `PlotView` and buttons for interacting with it."""

    def __init__(self):
        layout = QVBoxLayout()
        Widget.__init__(self, layout)
        self.view = PlotView()
        layout.addWidget(self.view)

        autoscale_button = Button("Autoscale", self.autoscale)
        Shortcut(self, "Ctrl+S", self.save_as_image)
        save_button = Button("Save", self.save_as_image)
        layout_util.add_to_layout(
            layout_util.add_sublayout(layout, QHBoxLayout()), autoscale_button, save_button
        )

    def save_as_image(self):
        file, _ = QFileDialog.getSaveFileName(
            None, "Save Graph", "untitled.png", "Portable Network Graphics (*.png)"
        )
        if file != "":
            try:
                self.view.plot_item.export_to_image(file)
            except Exception:
                errors.show_error("Save Error", "Failed to save graph.")

    def autoscale(self):
        """Autoscale the graph."""
        self.view.plot_item.getViewBox().enableAutoRange(pg.ViewBox.XYAxes)



from typing import Iterable

from ...utility.serde import Json
from . import tree_item
from .sequence_item import SequenceItem
from .tree_item import MutableTreeItem, TreeItem

ITEMS = "items"

type SubItemType = TreeItem[SequenceItem]


class RootItem[SubItem: TreeItem](MutableTreeItem[SubItem]):

    def __init__(self):
        self.subitems: list[SubItem] = []

    def serialize(self) -> list[Json]:  # implementation
        return [subitem.serialize() for subitem in self.subitems]

    def parent(self) -> TreeItem | None:  # implementation
        return None

    def set_parent(self, parent: TreeItem | None):
        if parent is not None:
            raise ValueError("The parent of a `RootItem` must always be `None`")

    def subitem_count(self) -> int:  # implementation
        return len(self.subitems)

    def index(self, item: TreeItem) -> int | None:  # implementation
        return tree_item.index(self.subitems, item)

    def get_subitem(self, index: int) -> SubItem | None:  # implementation
        return tree_item.get_subitem(self.subitems, index)

    def append_subitems(self, items: Iterable[SubItem]):  # implementation
        tree_item.append_subitems(self, self.subitems, items)

    def insert_subitems(self, start: int, items: Iterable[SubItem]):  # implementation
        tree_item.insert_subitems(self, self.subitems, start, items)

    def remove_subitems(self, start: int, count: int):  # implementation
        tree_item.remove_subitems(self.subitems, start, count)

    # debugging
    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {{ subitems: {self.subitems!r} }}"



from threading import RLock
from typing import Any


class DataLock[Data: Any]:
    def __init__(self, data: Data):
        self.lock = RLock()
        self.data = data

    def get(self) -> Data:
        """Get the inner data. Treat the returned value as immutable."""
        with self.lock:
            return self.data

    def set(self, data: Data):
        """Set the inner data to **data**."""
        with self.lock:
            self.data = data

    def __enter__(self) -> Data:
        self.lock.acquire()
        return self.data

    def __exit__(self, *args):  # we don't care about exceptions here
        self.lock.release()



from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QMainWindow, QSizePolicy, QWidget


class SecondaryWindow(QMainWindow):
    """
    Secondary window with `closed` signal.

    Parameters
    ----------
    title
        The window title text.
    central_widget
        The widget to set as the central widget (optional).
    parent
        This window's parent (optional).

    Notes
    -----
    If **parent** is specified, the created window will always be displayed on top of the parent.
    You must call `show()` on the window as normal.
    """

    closed = pyqtSignal()

    def __init__(
        self, title: str, central_widget: QWidget | None = None, parent: QWidget | None = None
    ):
        QMainWindow.__init__(self, parent)
        self.setWindowTitle(title)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setCentralWidget(central_widget)  # takes ownership
        if central_widget is not None:
            central_widget.setVisible(True)

    def closeEvent(self, event: QCloseEvent | None):  # overridden method
        if event is not None:
            self.closed.emit()
        QMainWindow.closeEvent(self, event)

    def close_silent(self):
        """Close the window without emitting the `closed` signal."""
        self.blockSignals(True)
        self.close()
        self.blockSignals(False)



import random
import time
from contextlib import AsyncExitStack
from pathlib import Path
from typing import Any

# the `StepRunner` class is how you interact with the GUI
from fabrial import LineSettings, PlotSettings, SequenceStep, StepRunner

# we import `NAME` so we always use the same "Random Data" string
from .random_widget import NAME


class RandomDataStep(SequenceStep):
    """Record random data on an interval; step."""

    def __init__(self, data_interval: float):
        # we store this for later
        self.data_interval = data_interval

    # notice the `async def` on this function. `run()` must be
    # asynchronous, otherwise the step will not run. This is a
    # common mistake, so make sure you use `async def`!
    async def run(self, runner: StepRunner, data_directory: Path):
        # an `AsyncExitStack()` can have items added
        # to it during the context block
        async with AsyncExitStack() as exit_stack:
            plot_handle = await exit_stack.enter_async_context(
                runner.create_plot(
                    self, "A Random Plot", PlotSettings("Random Data", "Time (s)", "Value")
                )
            )
            # creating a plot does not automatically create a line
            # `add_line()` is asynchronous
            line_handle = await plot_handle.add_line(
                LineSettings("The Data", "red", 3, "o", "blue", 5)
            )
            # we record the start time for plotting
            start_time = time.time()
            # `data_directory` is where our step should record data
            f = exit_stack.enter_context(open(data_directory.joinpath("random_data.txt"), "w"))
            for _ in range(20):  # we'll record 20 random data points
                data = random.random()
                f.write(f"{data}\n")
                # use the `add_point()` to plot new points
                line_handle.add_point(time.time() - start_time, data)
                # `self.sleep()` pauses our step and lets the sequence
                # do something else. Notice the `await` keyword.
                # `self.sleep()` is also asynchronous, so this is the
                # syntax we use to call it. There is also
                # `self.sleep_until()` that you can use with
                # `time.time()`.
                await self.sleep(self.data_interval)
        # at the end of the block, `plot_handle` and `line_handle` or no longer valid

        # we can send a prompt to the user with `prompt_user()`.
        # It will show buttons based on the supplied dictionary,
        # and it will return an integer corresponding to the pressed
        # button (i.e. 0 for "Happy" and 1 for "Sad")
        response = await runner.prompt_user(
            self, "We finished! How do we feel?", {0: "Happy", 1: "Sad"}
        )
        match response:
            case 0:  # happy
                print("You feel happy")
            case 1:  # sad
                print("You feel sad")

    def reset(self):
        # our step doesn't have anything to "reset". If your step
        # accessed a physical device, for example, this function
        # would clean up and disconnect from that device
        pass  # do nothing

    def name(self) -> str:
        return NAME

    def metadata(self) -> dict[str, Any]:
        return {"Selected Data Interval": self.data_interval}



from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout

from .....classes import DescriptionInfo
from .....constants.paths.process import filenames
from .....custom_widgets import SpinBox
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import BackgroundTemperatureProcess


class BackgroundTemperatureWidget(AbstractBaseWidget):
    """Enable background temperature monitoring."""

    DISPLAY_NAME_PREFIX = "Enable Background Temperature Monitoring"

    def __init__(self):
        layout = QFormLayout()
        self.interval_spinbox = SpinBox(50)
        AbstractBaseWidget.__init__(
            self,
            layout,
            self.DISPLAY_NAME_PREFIX,
            BackgroundTemperatureProcess,
            "system-monitor.png",
            DescriptionInfo(
                "temperature",
                "background_temperature",
                BackgroundTemperatureProcess.directory_name(),
                DescriptionInfo.Substitutions(
                    parameters_dict={
                        "MEASUREMENT_INTERVAL": encoding.MEASUREMENT_INTERVAL,
                        "MINIMUM_INTERVAL": str(self.interval_spinbox.minimum()),
                    },
                    data_recording_dict={"TEMPERATURE_FILE": filenames.TEMPERATURES},
                ),
            ),
        )
        layout.addRow(encoding.MEASUREMENT_INTERVAL, self.interval_spinbox)

    @classmethod
    def from_dict(cls, data_as_dict: dict[str, Any]) -> Self:
        widget = cls()
        widget.interval_spinbox.setValue(data_as_dict[encoding.MEASUREMENT_INTERVAL])
        return widget

    def to_dict(self) -> dict[str, Any]:
        return {encoding.MEASUREMENT_INTERVAL: self.interval_spinbox.value()}



import typing
from os import PathLike

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QTabWidget

from ..classes import DataLock, LineIndex, LineSettings, PlotIndex, PlotSettings, Shortcut
from ..custom_widgets import Button, PlotWidget
from ..secondary_window import SecondaryWindow
from ..utility import images


class SequenceDisplayTab(QTabWidget):
    """Tab for displaying graphing widgets during the sequence."""

    def __init__(self):
        QTabWidget.__init__(self)
        self.plots: list[PlotWidget] = []
        self.popped_graphs: dict[int, SecondaryWindow] = {}

        self.setCornerWidget(Button("Pop Graph", self.pop_graph), Qt.Corner.TopRightCorner)
        self.setMovable(True)  # allow moving tabs around
        Shortcut(self, "Ctrl+G", self.pop_graph)

        self.sequence_step_map: dict[int, tuple[QTabWidget, dict[int, PlotWidget]]] = {}

        self.step_tab_icon = images.make_icon("category.png")
        self.plot_tab_icon = images.make_icon("chart-up.png")

    def get_plot(self, plot_index: PlotIndex) -> PlotWidget:
        """
        Get the plot corresponding to **plot_index**.

        Raises
        ------
        IndexError
            The plot at **plot_index** does not exist. This is fatal.
        """
        # indexing goes
        # dict[int, tuple[QTabWidget, dict[int, PlotWidget]]]
        # tuple[QTabWidget, dict[int, PlotWidget]]
        # dict[int, PlotWidget]
        # PlotWidget
        return self.sequence_step_map[plot_index.step_address][1][plot_index.plot_number]

    def add_plot(
        self,
        step_address: int,
        step_name: str,
        tab_text: str,
        plot_settings: PlotSettings,
        receiver: DataLock[PlotIndex | None],
    ):
        """
        Create a new tab for the step at **step_address** (if there isn't one already), then create
        a plot in that tab. Sends a `PlotIndex` to the **receiver** that can be used to index the
        new plot later.
        """
        # get the plots map if it exists, otherwise create and add it
        try:
            step_tab_widget, plots = self.sequence_step_map[step_address]
        except KeyError:
            plots = {}
            step_tab_widget = QTabWidget()  # create a new tab widget for the plot
            step_tab_widget.setMovable(True)  # allow moving plot tabs around
            self.addTab(step_tab_widget, self.step_tab_icon, step_name)
            self.sequence_step_map[step_address] = (step_tab_widget, plots)

        # create and initialize the plot
        plot_widget = PlotWidget()
        plot_item = plot_widget.view.plot_item  # shortcut
        plot_item.set_title(plot_settings.title)
        plot_item.set_label("bottom", plot_settings.x_label)
        plot_item.set_label("left", plot_settings.y_label)
        # we can use the address because removing the plot from this widget is the same as deleting
        # the plot
        plot_number = id(plot_widget)
        plots[plot_number] = plot_widget
        # add the plot to the step's tab widget
        step_tab_widget.addTab(plot_widget, self.plot_tab_icon, tab_text)

        receiver.set(PlotIndex(step_address, plot_number))  # send the index to the receiver

    def remove_plot(self, plot_index: PlotIndex):
        """Remove and delete the plot at **plot_index**."""
        # remove the plot from the map
        plot_tab_widget, plot_map = self.sequence_step_map[plot_index.step_address]
        # remove the plot
        plot_widget = plot_map.pop(plot_index.plot_number)
        # un-pop the graph if it is popped
        if (popped_graph := self.popped_graphs.get(id(plot_widget))) is not None:
            popped_graph.close()
        plot_widget.setParent(None)
        plot_widget.deleteLater()
        # if there are no more plots, remove the entire tab
        if len(plot_map) == 0:
            self.sequence_step_map.pop(plot_index.step_address)  # remove from the map
            plot_tab_widget.setParent(None)
            plot_tab_widget.deleteLater()

    def add_line(
        self,
        plot_index: PlotIndex,
        line_settings: LineSettings,
        receiver: DataLock[LineIndex | None],
    ):
        """
        Add a new line to the plot at **plot_index** configured using **line_settings**. Sends a
        `LineIndex` to the **receiver** that can be used to index the new line later.
        """
        # create a new empty line
        plot_item = self.get_plot(plot_index).view.plot_item
        # we can use the line count as an index because you can't remove lines
        line_number = plot_item.line_count()  # store this before adding the new line
        plot_item.plot(
            [],
            [],
            line_settings.legend_label,
            line_settings.line_color,
            line_settings.line_width,
            line_settings.symbol,
            line_settings.symbol_color,
            line_settings.symbol_size,
        )
        receiver.set(LineIndex(plot_index, line_number))  # send the index to the receiver

    def add_point(self, line_index: LineIndex, x: float, y: float):
        """Add a point to the line at **line_index**."""
        self.get_plot(line_index.plot_index).view.plot_item.add_point(x, y, line_index.line_number)

    def set_log_scale(self, plot_index: PlotIndex, x_log: bool | None, y_log: bool | None):
        """Set the whether the plot at **plot_index** uses a log scale."""
        self.get_plot(plot_index).view.plot_item.setLogMode(x_log, y_log)

    def save_plot(self, plot_index: PlotIndex, file: PathLike[str] | str):
        """Save the plot at **plot_index** to **file**."""
        # TODO: figure out if a failure to create the image file is fatal
        self.get_plot(plot_index).view.plot_item.export_to_image(str(file))

    def pop_graph(self):
        """Pop the current graph into a secondary window."""
        # this cast is safe because we only ever add `QTabWidget`s to this widget
        step_tab_widget = typing.cast(QTabWidget, self.currentWidget())
        if step_tab_widget is not None:
            plot = step_tab_widget.currentWidget()
            if plot is not None:
                tab_title = step_tab_widget.tabText(step_tab_widget.currentIndex())
                popped_window = SecondaryWindow(tab_title, plot)
                Shortcut(popped_window, "Ctrl+G", popped_window.close)
                plot_address = id(plot)
                self.popped_graphs[plot_address] = popped_window
                # closing the popped graph puts it back in its original parent
                popped_window.closed.connect(lambda: self.popped_graphs.pop(plot_address))
                popped_window.closed.connect(
                    lambda: step_tab_widget.addTab(plot, self.plot_tab_icon, tab_title)
                )
                # show and ensure the initial size is correct
                popped_window.show()
                popped_window.resize(popped_window.sizeHint())



from typing import Any

from polars import DataFrame

from .....classes import AbstractForegroundProcess, ProcessRunner
from .....instruments import INSTRUMENTS
from .....utility import dataframe
from . import encoding


class SetNoStabilizeProcess(AbstractForegroundProcess):
    """Set the oven's setpoint without waiting to stabilize."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractForegroundProcess.__init__(self, runner, data, name)
        self.oven = INSTRUMENTS.oven
        self.setpoint = data[encoding.SETPOINT]

    @staticmethod
    def directory_name() -> str:
        return "Set Temperature (No Stabilization)"

    def run(self):
        # change the oven's setpoint and make sure it actually happens
        while not self.oven.change_setpoint(self.setpoint):
            if not self.wait(50, self.oven.is_connected):
                break

    def metadata(self) -> DataFrame:
        return dataframe.add_to_dataframe(
            AbstractForegroundProcess.metadata(self), {"Selected Setpoint": self.setpoint}
        )



from pathlib import Path

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QDoubleSpinBox, QFormLayout

from fabrial import ItemWidget

# need to create this first so the icon is valid
app = QApplication([])

NAME = "Random Data"
DIRECTORY = Path(__file__).parent
ICON = QIcon(str(DIRECTORY.joinpath("example_icon")))


class RandomDataWidget(ItemWidget):
    """Record random data on an interval; widget."""

    def __init__(self, data_interval: float):
        layout = QFormLayout()
        ItemWidget.__init__(self, layout, NAME, ICON, None)

        self.interval_spinbox = QDoubleSpinBox()
        self.interval_spinbox.setMinimum(0.1)
        self.interval_spinbox.setMaximum(1000)
        self.interval_spinbox.setValue(data_interval)
        layout.addRow("Data Interval", self.interval_spinbox)


widget = RandomDataWidget(10)
widget.show()
app.exec()



from __future__ import annotations

import typing
from typing import Iterable, Mapping, Self

from PyQt6.QtGui import QIcon

from ...classes import SequenceStep
from ...utility import serde
from ...utility.serde import Json
from ..data_item import DataItem
from . import tree_item
from .tree_item import MutableTreeItem, TreeItem

SUBITEMS = "subitems"
ITEM = "item"


class SequenceItem(MutableTreeItem["SequenceItem"]):
    """An item that represents a sequence step."""

    def __init__(self, parent: TreeItem | None, data_item: DataItem):
        self.parent_item = parent
        self.item = data_item
        self.bold = False
        self.subitems: list[SequenceItem] = []

    @classmethod
    def from_dict(cls, parent: TreeItem, item_as_dict: Mapping[str, Json]) -> Self:
        """Create the item from a JSON dictionary."""
        # deserialize the inner item
        inner_item: DataItem = serde.deserialize(
            typing.cast(Mapping[str, Json], item_as_dict[ITEM])
        )
        item = cls(parent, inner_item)  # make the outer item
        # recursively deserialize all subitems
        subitems = [
            cls.from_dict(item, subitem_as_dict)
            for subitem_as_dict in typing.cast(Iterable[Mapping[str, Json]], item_as_dict[SUBITEMS])
        ]
        item.append_subitems(subitems)  # add them to the outer item
        return item

    def serialize(self) -> dict[str, Json]:  # implementation
        return {
            ITEM: self.item.serialize_tagged(),
            SUBITEMS: [subitem.serialize() for subitem in self.subitems],
        }

    def parent(self) -> TreeItem | None:  # implementation
        return self.parent_item

    def set_parent(self, parent: TreeItem | None):
        self.parent_item = parent

    def display_name(self) -> str:  # overridden
        return self.item.display_name()

    def icon(self) -> QIcon:  # overridden
        return self.item.icon()

    def supports_subitems(self) -> bool:  # overridden
        return self.item.supports_subitems()

    def supports_dragging(self) -> bool:  # overridden
        return True

    def subitem_count(self) -> int:  # implementation
        return len(self.subitems)

    def index(self, item: TreeItem) -> int | None:  # implementation
        return tree_item.index(self.subitems, item)

    def get_subitem(self, index: int) -> SequenceItem | None:  # implementation
        return tree_item.get_subitem(self.subitems, index)

    def append_subitems(self, items: Iterable[SequenceItem]):  # implementation
        tree_item.append_subitems(self, self.subitems, items)

    def insert_subitems(self, start: int, items: Iterable[SequenceItem]):  # implementation
        tree_item.insert_subitems(self, self.subitems, start, items)

    def remove_subitems(self, start: int, count: int):  # implementation
        tree_item.remove_subitems(self.subitems, start, count)

    def open_event(self, editable: bool) -> bool:  # overridden
        self.item.open_event(editable)
        return False

    def expand_event(self):
        self.item.expand_event()

    def collapse_event(self):
        self.item.collapse_event()

    def set_bold(self, bold: bool):
        """Set whether the item is bolded."""
        self.bold = bold

    def is_bold(self) -> bool:
        """Whether the item is bolded."""
        return self.bold

    def create_sequence_step(self, substeps: Iterable[SequenceStep]) -> SequenceStep:
        """Create the `SequenceStep` this item represents."""
        return self.item.create_sequence_step(substeps)

    # debugging
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} {{ "
            f"bold: {self.is_bold()!r}, "
            f"item: {self.item!r}, "
            f"subitems: {self.subitems!r} }}"
        )



from PyQt6.QtWidgets import QFrame, QSizePolicy


class HSeparator(QFrame):
    """A horizontal separator."""

    def __init__(self):
        QFrame.__init__(self)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setLineWidth(1)
        self.setFrameShadow(QFrame.Shadow.Sunken)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)


class VSeparator(QFrame):
    """A vertical separator."""

    def __init__(self):
        QFrame.__init__(self)
        self.setFrameShape(QFrame.Shape.VLine)
        self.setLineWidth(1)
        self.setFrameShadow(QFrame.Shadow.Sunken)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum)



from typing import Any, Callable

from PyQt6.QtCore import QCoreApplication, QEventLoop, QTimer


def PROCESS_EVENTS():
    """Call `QCoreApplication.processEvents()` and let it run for 10ms."""
    QCoreApplication.processEvents(QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents, 10)


def delay_until_running(callback: Callable[[], Any]):
    """Queue a callback for the next run of the event loop."""
    QTimer.singleShot(0, callback)



import os
import sys
from types import ModuleType

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication
from tendo.singleton import SingleInstance, SingleInstanceException

from .constants import PACKAGE_NAME, icons
from .constants.paths import FOLDERS_TO_CREATE
from .gamry_integration import GAMRY
from .main_window import MainWindow
from .utility import errors, plugins


def make_application_folders():
    """Create folders for the application."""
    for folder in FOLDERS_TO_CREATE:
        os.makedirs(folder, exist_ok=True)


def check_for_other_instances() -> SingleInstance:
    """
    Attempt to create a singleton so only once instance of the application can run. If an instance
    is already running, this exits the application.
    """
    try:
        return SingleInstance()
    except SingleInstanceException:
        sys.exit()


def fix_windows_sucking():
    """Make the application show up in the taskbar on Windows (because Windows sucks)."""
    try:
        from ctypes import windll  # only exists on Windows

        appid = f"maughangroup.{PACKAGE_NAME}"
        windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)
    except ImportError:
        pass


def load_plugins() -> dict[str, ModuleType]:
    """Load all plugins."""
    plugin_modules, global_failures, local_failures = plugins.load_all_plugins()
    if len(global_failures) > 0 or len(local_failures) > 0:
        message = ""
        if len(global_failures) > 0:
            message += f"Failed to load global plugins:\n\n{", ".join(global_failures)}\n\n"
        if len(local_failures) > 0:
            message += f"Failed to load local plugins:\n\n{", ".join(local_failures)}\n\n"
        message += "See the error log for details."
        errors.show_error_delayed("Plugin Error", message)

    return plugin_modules


def main():
    me = check_for_other_instances()  # noqa
    errors.set_up_logging()
    sys.excepthook = errors.exception_handler  # log all uncaught exceptions
    fix_windows_sucking()
    make_application_folders()
    # suppress Qt warnings (there is a bug that generates warnings when the window is resized)
    errors.suppress_warnings()
    # make the app first
    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(str(icons.MAIN_ICON)))
    # then load plugins
    plugin_modules = load_plugins()
    # then make the main window
    main_window = MainWindow(plugin_modules)
    main_window.showMaximized()

    # app, main_window = make_app(plugin_modules)
    # run the application
    app.exec()
    # TODO: remove these
    # stop the oven's thread
    # close GamryCOM
    GAMRY.cleanup()

    # check to relaunch
    if main_window.should_relaunch():
        del me  # make sure the singleton is cleared
        # replace the current process with a new version of this one
        os.execl(sys.executable, sys.executable, *sys.argv)



from typing import Any, Callable, Self

from PyQt6.QtCore import QObject, QTimer


class Timer(QTimer):
    """
    Easier QTimer class. The passes in **slots** are automatically connected to the `timeout`
    signal.

    Parameters
    ----------
    parent
        The QObject that owns this timer.
    interval_ms
        The timeout interval in milliseconds.
    slots
        Function(s) to call when the timer times out.
    """

    def __init__(self, parent: QObject | None, intverval_ms: int, *slots: Callable[[], Any]):
        QTimer.__init__(self, parent)
        self.setInterval(intverval_ms)
        for slot in slots:
            self.timeout.connect(slot)

    def start_fast(self) -> Self:
        """Start the timer and instantly emit the timeout signal."""
        self.start()
        self.timeout.emit()
        return self



from typing import Any

import polars as pl


def add_to_dataframe(dataframe: pl.DataFrame, data_to_add: dict[Any, Any]) -> pl.DataFrame:
    """
    Create a DataFrame from **data_to_add** and horizontally concatenate it with **dataframe**.
    The original dataframe's data is concatenated *to the right* of the new data.
    """
    return pl.concat((pl.DataFrame(data_to_add), dataframe), how="horizontal")



from typing import Any

from .....classes import AbstractForegroundProcess, ProcessRunner
from .....utility import dataframe
from . import encoding


class LoopProcess(AbstractForegroundProcess):
    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractForegroundProcess.__init__(self, runner, data, name)
        self.number_of_loops = data[encoding.NUMBER_OF_LOOPS]
        self.loops_completed = 0

    def run(self) -> None:
        # reconfigure the runner while saving previous values
        runner = self.runner()
        previous_directory = runner.directory()
        previous_number = runner.number()
        runner.set_directory(self.directory())
        runner.reset_number()

        # run each item and repeat for the loop number
        loop_item = runner.current_item()
        if loop_item is not None:
            while self.loops_completed < self.number_of_loops:
                for item in loop_item.subitems():
                    process_type = item.process_type()
                    if process_type is not None:
                        widget = item.widget()
                        process = process_type(runner, widget.to_dict(), widget.display_name())
                        if not self.process_runner.run_process(process, item):
                            break
                else:  # https://www.geeksforgeeks.org/python-for-else/ (fucking crazy right??)
                    self.loops_completed += 1
                    continue
                break
        # cleanup
        runner.set_directory(previous_directory)
        runner.set_number(previous_number)

    @staticmethod
    def directory_name():
        return "Loop"

    def metadata(self):
        return dataframe.add_to_dataframe(
            AbstractForegroundProcess.metadata(self),
            {
                "Selected Loop Count": self.number_of_loops,
                "Actual Loops Completed": self.loops_completed,
            },
        )



from PyQt6.QtWidgets import QCheckBox, QFormLayout

from ...constants.paths import settings
from ..augmented import Widget


class SequenceSettingsTab(Widget):
    """Settings related to the sequence."""

    def __init__(self):
        layout = QFormLayout()
        Widget.__init__(self, layout)
        self.non_empty_directory_warning_checkbox = QCheckBox(
            "Show a warning when starting the sequence with a non-empty data directory."
        )
        try:
            with open(settings.sequence.NON_EMPTY_DIRECTORY_WARNING_FILE, "r") as f:
                if f.read().strip() == str(True):
                    self.non_empty_directory_warning_checkbox.setChecked(True)
        except Exception:
            self.non_empty_directory_warning_checkbox.setChecked(True)

        layout.addWidget(self.non_empty_directory_warning_checkbox)

    def save_on_close(self):
        """Call this when closing the settings window to save settings."""
        with open(settings.sequence.NON_EMPTY_DIRECTORY_WARNING_FILE, "w") as f:
            f.write(str(self.non_empty_directory_warning_checkbox.isChecked()))



from __future__ import annotations

import asyncio
import copy
from dataclasses import dataclass
from os import PathLike
from typing import TYPE_CHECKING

from pyqtgraph import PlotDataItem

from .lock import DataLock

if TYPE_CHECKING:
    from .step_runner import StepRunner


class LineData:
    """Container for a line and its data. This is similar to a `Line2D` in `matplotlib`."""

    def __init__(self, line: PlotDataItem, x_data: list[float], y_data: list[float]):
        self.line = line
        self.x_data = x_data
        self.y_data = y_data

    def add_point(self, x: float, y: float):
        """Add a point to the line."""
        self.x_data.append(x)
        self.y_data.append(y)
        self.line.setData(self.x_data, self.y_data)


@dataclass
class PlotSettings:
    """
    Container for plot settings (i.e. title and axis labels).

    Parameters
    ----------
    title
        The plot's title.
    x_label
        The plot's x-label.
    y_label
        The plot's y-label.
    """

    title: str
    x_label: str
    y_label: str

    def __copy__(self) -> PlotSettings:
        return PlotSettings(self.title, self.x_label, self.y_label)


@dataclass
class LineSettings:
    """
    Container for line settings (i.e. the line width, color, etc.).

    Parameters
    ----------
    legend_label
        The label to use for this line in the legend.
    line_color
        The line's color (for example, "red", or "#112233" for an exact color). If `None`, there
        will only be markers.
    line_width
        The line's width. If `None`, there will only be markers.
    symbol
        The marker symbol, i.e. "o" for a dot.
    symbol_size
        The point size.
    symbol_color
        The color of the points. Same style as **line_color**.
    """

    legend_label: str
    line_color: str | None
    line_width: float | None
    symbol: str
    symbol_color: str
    symbol_size: int

    def __copy__(self) -> LineSettings:
        return LineSettings(
            self.legend_label,
            self.line_color,
            self.line_width,
            self.symbol,
            self.symbol_color,
            self.symbol_size,
        )


@dataclass
class PlotIndex:
    """
    An index to a plot on the visuals tab.

    Parameters
    ----------
    step_address
        The memory address (aka the result of `id()`) of the step that created the plot.
    plot_number
        The number that can be used to index the actual plot.
    """

    step_address: int
    plot_number: int

    def __copy__(self) -> PlotIndex:
        return PlotIndex(self.step_address, self.plot_number)


@dataclass
class LineIndex:
    """
    An index to a line on a plot.

    Parameters
    ----------
    plot_index
        The `PlotIndex` of the plot where this the line exists.
    line_number
        The number that can be used to index the actual line.
    """

    plot_index: PlotIndex
    line_number: int

    def __copy__(self) -> LineIndex:
        return LineIndex(copy.copy(self.plot_index), self.line_number)


class PlotHandle:
    """
    A thread-safe handle to a plot on the sequence visuals tab.

    Parameters
    ----------
    runner
        The `StepRunner` being used by the sequence.
    plot_index
        The `PlotIndex` that can be used to index the actual plot.
    """

    def __init__(self, runner: StepRunner, plot_index: PlotIndex):
        self.runner = runner
        self.plot_index = plot_index

    def set_log_scale(self, x_log: bool | None, y_log: bool | None):
        """
        Set whether the x- and/or y-axis use a logarithmic scale. A value of `None` for **x_log** or
        **y_log** will leave the corresponding axis unchanged.
        """
        plot_index = copy.copy(self.plot_index)
        self.runner.submit_plot_command(
            lambda plot_tab: plot_tab.set_log_scale(plot_index, x_log, y_log)
        )

    def save_plot(self, file: PathLike[str] | str):
        """Save the plot to **file**."""
        plot_index = copy.copy(self.plot_index)
        self.runner.submit_plot_command(lambda plot_tab: plot_tab.save_plot(plot_index, file))

    async def add_line(self, line_settings: LineSettings) -> LineHandle:
        """Add an empty line to the plot."""
        receiver: DataLock[LineIndex | None] = DataLock(None)
        # we make copies of the plot index because sending the original is not thread-safe
        plot_index = copy.copy(self.plot_index)
        self.runner.submit_plot_command(
            lambda plot_tab: plot_tab.add_line(plot_index, line_settings, receiver)
        )
        while True:
            await asyncio.sleep(0)
            if (line_index := receiver.get()) is not None:
                return LineHandle(self, line_index)


class LineHandle:
    """
    A thread-safe handle to a line on a plot.

    Parameters
    ----------
    plot_handle
        The `PlotHandle` that created this object.
    line_index
        The `LineIndex` that can be used to index the actual line.
    """

    def __init__(self, plot_handle: PlotHandle, line_index: LineIndex):
        self.parent = plot_handle
        self.line_index = line_index

    def add_point(self, x: float, y: float):
        """Add a point to the line."""
        line_index = copy.copy(self.line_index)
        self.parent.runner.submit_plot_command(
            lambda plot_tab: plot_tab.add_point(line_index, x, y)
        )
